import os
from typing import Optional

import pandas as pd
import torch

from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.preprocessing.datasets.chebi import _ChEBIDataExtractor
from chebai.preprocessing.datasets.pubchem import LabeledUnlabeledMixed


class BCEWeighted(torch.nn.BCEWithLogitsLoss):
    """
    BCEWithLogitsLoss with weights automatically computed according to the beta parameter.
    If beta is None or data_extractor is None, the loss is unweighted.

    This class computes weights based on the formula from the paper by Cui et al. (2019):
    https://openaccess.thecvf.com/content_CVPR_2019/papers/Cui_Class-Balanced_Loss_Based_on_Effective_Number_of_Samples_CVPR_2019_paper.pdf

    Args:
        beta (float, optional): The beta parameter for weight calculation. Default is None.
        data_extractor (XYBaseDataModule, optional): The data extractor for loading the dataset. Default is None.
    """

    def __init__(
        self,
        beta: Optional[float] = None,
        data_extractor: Optional[XYBaseDataModule] = None,
        **kwargs,
    ):
        self.beta = beta
        if isinstance(data_extractor, LabeledUnlabeledMixed):
            data_extractor = data_extractor.labeled
        self.data_extractor = data_extractor
        assert (
            isinstance(self.data_extractor, _ChEBIDataExtractor)
            or self.data_extractor is None
        )
        super().__init__(**kwargs)

    def set_pos_weight(self, input: torch.Tensor) -> None:
        """
        Sets the positive weights for the loss function based on the input tensor.

        Args:
            input (torch.Tensor): The input tensor for which to set the positive weights.
        """
        if (
            self.beta is not None
            and self.data_extractor is not None
            and all(
                os.path.exists(
                    os.path.join(self.data_extractor.processed_dir, file_name)
                )
                for file_name in self.data_extractor.processed_file_names
            )
            and self.pos_weight is None
        ):
            print(
                f"Computing loss-weights based on v{self.data_extractor.chebi_version} dataset (beta={self.beta})"
            )
            complete_labels = torch.concat(
                [
                    torch.stack(
                        [
                            torch.Tensor(row["labels"])
                            for row in self.data_extractor.load_processed_data(
                                filename=file_name
                            )
                        ]
                    )
                    for file_name in self.data_extractor.processed_file_names
                ]
            )
            value_counts = complete_labels.sum(dim=0)
            weights = [
                (1 - self.beta) / (1 - pow(self.beta, value)) for value in value_counts
            ]
            mean = sum(weights) / len(weights)
            self.pos_weight = torch.tensor(
                [w / mean for w in weights], device=input.device
            )

    def forward(
        self, input: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """
        Forward pass for the loss calculation.

        Args:
            input (torch.Tensor): The input tensor (predictions).
            target (torch.Tensor): The target tensor (labels).

        Returns:
            torch.Tensor: The computed loss.
        """
        self.set_pos_weight(input)
        return super().forward(input, target)



from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Dict, List, Set, Tuple

import networkx as nx
import pandas as pd


class MockOntologyGraphData(ABC):
    """
    Abstract base class for mocking ontology graph data.

    This class provides a set of static methods that must be implemented by subclasses
    to return various elements of an ontology graph such as nodes, edges, and dataframes.
    """

    @staticmethod
    @abstractmethod
    def get_nodes() -> List[int]:
        """
        Get a list of node IDs in the ontology graph.

        Returns:
            List[int]: A list of node IDs.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_number_of_nodes() -> int:
        """
        Get the number of nodes in the ontology graph.

        Returns:
            int: The total number of nodes.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_edges() -> Set[Tuple[int, int]]:
        """
        Get the set of edges in the ontology graph.

        Returns:
            Set[Tuple[int, int]]: A set of tuples where each tuple represents an edge between two nodes.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_number_of_edges() -> int:
        """
        Get the number of edges in the ontology graph.

        Returns:
            int: The total number of edges.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_edges_of_transitive_closure_graph() -> Set[Tuple[int, int]]:
        """
        Get the set of edges in the transitive closure of the ontology graph.

        Returns:
            Set[Tuple[int, int]]: A set of tuples representing the transitive closure edges.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_number_of_transitive_edges() -> int:
        """
        Get the number of edges in the transitive closure of the ontology graph.

        Returns:
            int: The total number of transitive edges.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_obsolete_nodes_ids() -> Set[int]:
        """
        Get the set of obsolete node IDs in the ontology graph.

        Returns:
            Set[int]: A set of obsolete node IDs.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_transitively_closed_graph() -> nx.DiGraph:
        """
        Get the transitive closure of the ontology graph.

        Returns:
            nx.DiGraph: A directed graph representing the transitive closure of the ontology graph.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_data_in_dataframe() -> pd.DataFrame:
        """
        Get the ontology data as a Pandas DataFrame.

        Returns:
            pd.DataFrame: A DataFrame containing ontology data.
        """
        pass


class ChebiMockOntology(MockOntologyGraphData):
    """
    A mock ontology representing a simplified ChEBI (Chemical Entities of Biological Interest) structure.
    This class is used for testing purposes and includes nodes and edges representing chemical compounds
    and their relationships in a graph structure.

    Nodes:
    - CHEBI:12345 (Compound A)
    - CHEBI:54321 (Compound B)
    - CHEBI:67890 (Compound C)
    - CHEBI:11111 (Compound D)
    - CHEBI:22222 (Compound E)
    - CHEBI:99999 (Compound F)
    - CHEBI:77533 (Compound G, Obsolete node)
    - CHEBI:77564 (Compound H, Obsolete node)
    - CHEBI:88888 (Compound I)

    Valid Edges:
    - CHEBI:54321 -> CHEBI:12345
    - CHEBI:67890 -> CHEBI:12345
    - CHEBI:67890 -> CHEBI:88888
    - CHEBI:11111 -> CHEBI:54321
    - CHEBI:22222 -> CHEBI:67890
    - CHEBI:12345 -> CHEBI:99999

    The class also includes methods to retrieve nodes, edges, and transitive closure of the graph.

    Visual Representation Graph with Valid Nodes and Edges:

                                       22222
                                        /
                       11111         67890
                         \\         /  \
                        54321     /    88888
                           \\   /
                           12345
                             \
                            99999
    """

    @staticmethod
    def get_nodes() -> List[int]:
        """
        Get the set of valid node IDs in the mock ontology.

        Returns:
        - Set[int]: A set of integers representing the valid ChEBI node IDs.
        """
        return [11111, 12345, 22222, 54321, 67890, 88888, 99999]

    @staticmethod
    def get_number_of_nodes() -> int:
        """
        Get the number of valid nodes in the mock ontology.

        Returns:
        - int: The number of valid nodes.
        """
        return len(ChebiMockOntology.get_nodes())

    @staticmethod
    def get_edges() -> Set[Tuple[int, int]]:
        """
        Get the set of valid edges in the mock ontology.

        Returns:
        - Set[Tuple[int, int]]: A set of tuples representing the directed edges
          between ChEBI nodes.
        """
        return {
            (54321, 12345),
            (67890, 12345),
            (67890, 88888),
            (11111, 54321),
            (22222, 67890),
            (12345, 99999),
        }

    @staticmethod
    def get_number_of_edges() -> int:
        """
        Get the number of valid edges in the mock ontology.

        Returns:
        - int: The number of valid edges.
        """
        return len(ChebiMockOntology.get_edges())

    @staticmethod
    def get_edges_of_transitive_closure_graph() -> Set[Tuple[int, int]]:
        """
        Get the set of edges derived from the transitive closure of the mock ontology graph.

        Returns:
        - Set[Tuple[int, int]]: A set of tuples representing the directed edges
          in the transitive closure of the ChEBI graph.
        """
        return {
            (54321, 12345),
            (54321, 99999),
            (67890, 12345),
            (67890, 99999),
            (67890, 88888),
            (11111, 54321),
            (11111, 12345),
            (11111, 99999),
            (22222, 67890),
            (22222, 12345),
            (22222, 99999),
            (22222, 88888),
            (12345, 99999),
        }

    @staticmethod
    def get_number_of_transitive_edges() -> int:
        """
        Get the number of edges in the transitive closure of the mock ontology graph.

        Returns:
        - int: The number of edges in the transitive closure graph.
        """
        return len(ChebiMockOntology.get_edges_of_transitive_closure_graph())

    @staticmethod
    def get_obsolete_nodes_ids() -> Set[int]:
        """
        Get the set of obsolete node IDs in the mock ontology.

        Returns:
        - Set[int]: A set of integers representing the obsolete ChEBI node IDs.
        """
        return {77533, 77564}

    @staticmethod
    def get_raw_data() -> str:
        """
        Get the raw data representing the mock ontology in OBO format.

        Returns:
        - str: A string containing the raw OBO data for the mock ChEBI terms.
        """
        return """
        [Term]
        id: CHEBI:12345
        name: Compound A
        subset: 2_STAR
        property_value: http://purl.obolibrary.org/obo/chebi/formula "C26H35ClN4O6S" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/charge "0" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/monoisotopicmass "566.19658" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/mass "567.099" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/inchikey "ROXPMFGZZQEKHB-IUKKYPGJSA-N" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1" xsd:string
        property_value: http://purl.obolibrary.org/obo/chebi/inchi "InChI=1S/C26H35ClN4O6S/c1-16(2)28-26(34)30(5)14-23-17(3)13-31(18(4)15-32)25(33)21-7-6-8-22(24(21)37-23)29-38(35,36)20-11-9-19(27)10-12-20/h6-12,16-18,23,29,32H,13-15H2,1-5H3,(H,28,34)/t17-,18-,23+/m0/s1" xsd:string
        xref: LINCS:LSM-20139
        is_a: CHEBI:54321
        is_a: CHEBI:67890

        [Term]
        id: CHEBI:54321
        name: Compound B
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1O" xsd:string
        is_a: CHEBI:11111
        is_a: CHEBI:77564

        [Term]
        id: CHEBI:67890
        name: Compound C
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1N" xsd:string
        is_a: CHEBI:22222

        [Term]
        id: CHEBI:11111
        name: Compound D
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1F" xsd:string

        [Term]
        id: CHEBI:22222
        name: Compound E
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1Cl" xsd:string

        [Term]
        id: CHEBI:99999
        name: Compound F
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1Br" xsd:string
        is_a: CHEBI:12345

        [Term]
        id: CHEBI:77533
        name: Compound G
        is_a: CHEBI:99999
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=C1Br" xsd:string
        is_obsolete: true

        [Term]
        id: CHEBI:77564
        name: Compound H
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "CC=C1Br" xsd:string
        is_obsolete: true

        [Typedef]
        id: has_major_microspecies_at_pH_7_3
        name: has major microspecies at pH 7.3
        is_cyclic: true
        is_transitive: false

        [Term]
        id: CHEBI:88888
        name: Compound I
        property_value: http://purl.obolibrary.org/obo/chebi/smiles "C1=CC=CC=C1[Mg+]" xsd:string
        is_a: CHEBI:67890
        """

    @staticmethod
    def get_data_in_dataframe() -> pd.DataFrame:
        data = OrderedDict(
            id=[
                12345,
                54321,
                67890,
                11111,
                22222,
                99999,
                88888,
            ],
            name=[
                "Compound A",
                "Compound B",
                "Compound C",
                "Compound D",
                "Compound E",
                "Compound F",
                "Compound I",
            ],
            SMILES=[
                "C1=CC=CC=C1",
                "C1=CC=CC=C1O",
                "C1=CC=CC=C1N",
                "C1=CC=CC=C1F",
                "C1=CC=CC=C1Cl",
                "C1=CC=CC=C1Br",
                "C1=CC=CC=C1[Mg+]",
            ],
            **{
                # -row- [12345, 54321, 67890, 11111, 22222, 99999, 88888]
                11111: [True, True, False, True, False, True, False],
                12345: [True, False, False, False, False, True, False],
                22222: [True, False, True, False, True, True, True],
                54321: [True, True, False, False, False, True, False],
                67890: [True, False, True, False, False, True, True],
                88888: [False, False, False, False, False, False, True],
                99999: [False, False, False, False, False, True, False],
            },
        )

        data_df = pd.DataFrame(data)

        # ------------- Code Approach -------
        # ancestors_of_nodes = {}
        # for parent, child in ChebiMockOntology.get_edges_of_transitive_closure_graph():
        #     if child not in ancestors_of_nodes:
        #         ancestors_of_nodes[child] = set()
        #     if parent not in ancestors_of_nodes:
        #         ancestors_of_nodes[parent] = set()
        #     ancestors_of_nodes[child].add(parent)
        #     ancestors_of_nodes[child].add(child)
        #
        # # For each node in the ontology, create a column to check if it's an ancestor of any other node or itself
        # for node in ChebiMockOntology.get_nodes():
        #     data_df[node] = data_df['id'].apply(
        #         lambda x: (x == node) or (node in ancestors_of_nodes[x])
        #     )

        return data_df

    @staticmethod
    def get_transitively_closed_graph() -> nx.DiGraph:
        """
        Create a directed graph, compute its transitive closure, and return it.

        Returns:
            g (nx.DiGraph): A transitively closed directed graph.
        """
        g = nx.DiGraph()

        for node in ChebiMockOntology.get_nodes():
            g.add_node(node, **{"smiles": "test_smiles_placeholder"})

        g.add_edges_from(ChebiMockOntology.get_edges_of_transitive_closure_graph())

        return g



import logging
import os
import sys

import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.metrics import F1
from sklearn.metrics import f1_score
from torch import nn
from torch_geometric import nn as tgnn
from torch_geometric.data import DataLoader

from data import ClassificationData, JCIClassificationData

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class PartOfNet(pl.LightningModule):
    def __init__(self, in_length, loops=10):
        super().__init__()
        self.loops = loops
        self.left_graph_net = tgnn.GATConv(in_length, in_length)
        self.right_graph_net = tgnn.GATConv(in_length, in_length)
        self.attention = nn.Linear(in_length, 1)
        self.global_attention = tgnn.GlobalAttention(self.attention)
        self.output_net = nn.Sequential(
            nn.Linear(2 * in_length, 2 * in_length),
            nn.Linear(2 * in_length, in_length),
            nn.Linear(in_length, 500),
        )
        self.f1 = F1(1, threshold=0.5)

    def _execute(self, batch, batch_idx):
        pred = self(batch)
        loss = F.binary_cross_entropy_with_logits(pred, batch.label)
        f1 = self.f1(batch.label, torch.sigmoid(pred))
        return loss, f1

    def training_step(self, *args, **kwargs):
        loss, f1 = self._execute(*args, **kwargs)
        self.log(
            "train_loss",
            loss.detach().item(),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "train_f1",
            f1.item(),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        return loss

    def validation_step(self, *args, **kwargs):
        with torch.no_grad():
            loss, f1 = self._execute(*args, **kwargs)
            self.log(
                "val_loss",
                loss.detach().item(),
                on_step=True,
                on_epoch=True,
                prog_bar=True,
                logger=True,
            )
            self.log(
                "val_f1",
                f1.item(),
                on_step=True,
                on_epoch=True,
                prog_bar=True,
                logger=True,
            )
            return loss

    def forward(self, x):
        a = self.left_graph_net(x.x_s, x.edge_index_s.long())
        b = self.right_graph_net(x.x_t, x.edge_index_t.long())
        return self.output_net(
            torch.cat(
                [
                    self.global_attention(a, x.x_s_batch),
                    self.global_attention(b, x.x_t_batch),
                ],
                dim=1,
            )
        )

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters())
        return optimizer


class JCINet(pl.LightningModule):
    def __init__(self, in_length, hidden_length, num_classes, loops=10):
        super().__init__()
        self.loops = loops

        self.node_net = nn.Sequential(
            nn.Linear(self.loops * in_length, hidden_length), nn.ReLU()
        )
        self.embedding = torch.nn.Embedding(800, in_length)
        self.left_graph_net = tgnn.GATConv(in_length, in_length, dropout=0.1)
        self.final_graph_net = tgnn.GATConv(in_length, hidden_length, dropout=0.1)
        self.attention = nn.Linear(hidden_length, 1)
        self.global_attention = tgnn.GlobalAttention(self.attention)
        self.output_net = nn.Sequential(
            nn.Linear(hidden_length, hidden_length),
            nn.Linear(hidden_length, num_classes),
        )
        self.f1 = F1(num_classes, threshold=0.5)

    def _execute(self, batch, batch_idx):
        pred = self(batch)
        labels = batch.label.float()
        loss = F.binary_cross_entropy_with_logits(pred, labels)
        f1 = f1_score(
            labels.cpu() > 0.5, torch.sigmoid(pred).cpu() > 0.5, average="micro"
        )
        return loss, f1

    def training_step(self, *args, **kwargs):
        loss, f1 = self._execute(*args, **kwargs)
        self.log(
            "train_loss",
            loss.detach().item(),
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "train_f1",
            f1.item(),
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        return loss

    def validation_step(self, *args, **kwargs):
        with torch.no_grad():
            loss, f1 = self._execute(*args, **kwargs)
            self.log(
                "val_loss",
                loss.detach().item(),
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                logger=True,
            )
            self.log(
                "val_f1",
                f1.item(),
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                logger=True,
            )
            return loss

    def forward(self, x):
        a = self.embedding(x.x)
        l = []
        for _ in range(self.loops):
            a = self.left_graph_net(a, x.edge_index.long())
            l.append(a)
        at = self.global_attention(self.node_net(torch.cat(l, dim=1)), x.x_batch)
        return self.output_net(at)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters())
        return optimizer


def train(train_loader, validation_loader):
    if torch.cuda.is_available():
        trainer_kwargs = dict(gpus=-1, accelerator="ddp")
    else:
        trainer_kwargs = dict(gpus=0)
    net = JCINet(100, 100, 500)
    tb_logger = pl_loggers.CSVLogger("../../logs/")
    checkpoint_callback = ModelCheckpoint(
        dirpath=os.path.join(tb_logger.log_dir, "checkpoints"),
        filename="{epoch}-{step}-{val_loss:.7f}",
        save_top_k=5,
        save_last=True,
        verbose=True,
        monitor="val_loss",
        mode="min",
    )
    trainer = pl.Trainer(
        logger=tb_logger,
        callbacks=[checkpoint_callback],
        replace_sampler_ddp=False,
        **trainer_kwargs
    )
    trainer.fit(net, train_loader, val_dataloaders=validation_loader)


if __name__ == "__main__":
    batch_size = int(sys.argv[1])
    # vl = ClassificationData("data/full_chebi", split="validation")
    # tr = ClassificationData("data/full_chebi", split="train")
    tr = JCIClassificationData("data/JCI_data", split="train")
    vl = JCIClassificationData("data/JCI_data", split="validation")

    train_loader = DataLoader(
        tr,
        shuffle=True,
        batch_size=batch_size,
        follow_batch=["x", "edge_index", "label"],
    )
    validation_loader = DataLoader(
        vl, batch_size=batch_size, follow_batch=["x", "edge_index", "label"]
    )

    train(train_loader, validation_loader)



import json
from pathlib import Path

import torch
from jsonargparse import CLI
from sklearn.metrics import multilabel_confusion_matrix

from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.result.utils import (
    load_data_instance,
    load_model_for_inference,
    parse_config_file,
)


class ClassesPropertiesGenerator:
    """
    Computes PPV (Positive Predictive Value) and NPV (Negative Predictive Value) and counts the number of
    true positives (TP), false positives (FP), true negatives (TN), and false negatives (FN)
    for each class in a multi-label classification problem using a PyTorch Lightning model.
    """

    @staticmethod
    def load_class_labels(path: Path) -> list[str]:
        """
        Load a list of class names from a .json or .txt file.

        Args:
            path: Path to the class labels file (txt or json).

        Returns:
            A list of class names, one per line.
        """
        path = Path(path)
        with path.open() as f:
            return [line.strip() for line in f if line.strip()]

    @staticmethod
    def compute_classwise_scores(
        y_true: list[torch.Tensor],
        y_pred: list[torch.Tensor],
        raw_preds: torch.Tensor,
        class_names: list[str],
    ) -> dict[str, dict[str, float]]:
        """
        Compute PPV (precision, TP/(TP+FP)), NPV (TN/(TN+FN)) and the number of TNs, FPs, FNs and TPs for each class
        in a multi-label setting.

        Args:
            y_true: List of binary ground-truth label tensors, one tensor per sample.
            y_pred: List of binary prediction tensors, one tensor per sample.
            class_names: Ordered list of class names corresponding to class indices.

        Returns:
            Dictionary mapping each class name to its PPV and NPV metrics:
            {
                "class_name": {"PPV": float, "NPV": float, "TN": int, "FP": int, "FN": int, "TP": int},
                ...
            }
        """
        # Stack per-sample tensors into (n_samples, n_classes) numpy arrays
        true_np = torch.stack(y_true).cpu().numpy().astype(int)
        pred_np = torch.stack(y_pred).cpu().numpy().astype(int)

        # Compute confusion matrix for each class
        cm = multilabel_confusion_matrix(true_np, pred_np)

        results: dict[str, dict[str, float]] = {}
        for idx, cls_name in enumerate(class_names):
            tn, fp, fn, tp = cm[idx].ravel()
            tpv = tp / (tp + fp) if (tp + fp) > 0 else 0.0
            npv = tn / (tn + fn) if (tn + fn) > 0 else 0.0
            # positive_raw = [p.item() for i, p in enumerate(raw_preds[:, idx]) if true_np[i, idx]]
            # negative_raw = [p.item() for i, p in enumerate(raw_preds[:, idx]) if not true_np[i, idx]]
            results[cls_name] = {
                "PPV": round(tpv, 4),
                "NPV": round(npv, 4),
                "TN": int(tn),
                "FP": int(fp),
                "FN": int(fn),
                "TP": int(tp),
                # "positive_preds": positive_raw,
                # "negative_preds": negative_raw,
            }
        return results

    def generate_props(
        self,
        model_ckpt_path: str,
        model_config_file_path: str,
        data_config_file_path: str,
        output_path: str | None = None,
    ) -> None:
        """
        Run inference on validation set, compute TPV/NPV  per class, and save to JSON.

        Args:
            model_ckpt_path: Path to the PyTorch Lightning checkpoint file.
            model_config_file_path: Path to yaml config file of the model.
            data_config_file_path: Path to yaml config file of the data.
            output_path: Optional path where to write the JSON metrics file.
                         Defaults to '<processed_dir_main>/classes.json'.
        """
        print("Extracting validation data for computation...")

        data_cls_path, data_cls_kwargs = parse_config_file(data_config_file_path)
        data_module: XYBaseDataModule = load_data_instance(
            data_cls_path, data_cls_kwargs
        )

        splits_file_path = Path(data_module.processed_dir_main, "splits.csv")
        if data_module.splits_file_path is None:
            if not splits_file_path.exists():
                raise RuntimeError(
                    "Either the data module should be initialized with a `splits_file_path`, "
                    f"or the file `{splits_file_path}` must exists.\n"
                    "This is to prevent the data module from dynamically generating the splits."
                )

            print(
                f"`splits_file_path` is not provided as an initialization parameter to the data module\n"
                f"Using splits from the file {splits_file_path}"
            )
            data_module.splits_file_path = splits_file_path

        model_class_path, model_kwargs = parse_config_file(model_config_file_path)
        model = load_model_for_inference(
            model_ckpt_path, model_class_path, model_kwargs
        )

        val_loader = data_module.val_dataloader()
        print("Running inference on validation data...")

        y_true, y_pred = [], []
        raw_preds = []
        for batch_idx, batch in enumerate(val_loader):
            data = model._process_batch(  # pylint: disable=W0212
                batch, batch_idx=batch_idx
            )
            labels = data["labels"]
            outputs = model(data, **data.get("model_kwargs", {}))
            logits = outputs["logits"] if isinstance(outputs, dict) else outputs
            preds = torch.sigmoid(logits) > 0.5
            y_pred.extend(preds)
            y_true.extend(labels)
            raw_preds.extend(torch.sigmoid(logits))
        raw_preds = torch.stack(raw_preds)
        print("Computing metrics...")
        classes_file = Path(data_module.processed_dir_main) / "classes.txt"
        if output_path is None:
            output_file = Path(data_module.processed_dir_main) / "classes.json"
        else:
            output_file = Path(output_path)

        class_names = self.load_class_labels(classes_file)
        metrics = self.compute_classwise_scores(y_true, y_pred, raw_preds, class_names)

        with output_file.open("w") as f:
            json.dump(metrics, f, indent=2)
        print(f"Saved metrics to {output_file}")


class Main:
    """
    CLI wrapper for ClassesPropertiesGenerator.
    """

    def generate(
        self,
        model_ckpt_path: str,
        model_config_file_path: str,
        data_config_file_path: str,
        output_path: str | None = None,
    ) -> None:
        """
        CLI command to generate JSON with metrics on validation set.

        Args:
            model_ckpt_path: Path to the PyTorch Lightning checkpoint file.
            model_config_file_path: Path to yaml config file of the model.
            data_config_file_path: Path to yaml config file of the data.
            output_path: Optional path where to write the JSON metrics file.
                         Defaults to '<processed_dir_main>/classes.json'.
        """
        generator = ClassesPropertiesGenerator()
        generator.generate_props(
            model_ckpt_path,
            model_config_file_path,
            data_config_file_path,
            output_path,
        )


if __name__ == "__main__":
    # _generate_classes_props_json.py generate \
    # --model_ckpt_path "model/ckpt/path" \
    # --model_config_file_path "model/config/file/path" \
    # --data_config_file_path "data/config/file/path" \
    # --output_path "output/file/path" # Optional
    CLI(Main, as_positional=False)



import unittest
from unittest.mock import MagicMock, PropertyMock, patch

from chebai.preprocessing.datasets.base import XYBaseDataModule


class TestXYBaseDataModule(unittest.TestCase):
    """
    Unit tests for the methods of the XYBaseDataModule class.
    """

    @classmethod
    @patch.object(XYBaseDataModule, "_name", new_callable=PropertyMock)
    @patch("os.makedirs", return_value=None)
    def setUpClass(cls, mock_makedirs, mock_name_property: PropertyMock) -> None:
        """
        Set up a base instance of XYBaseDataModule for testing.
        """

        # Mock the _name property of XYBaseDataModule
        mock_name_property.return_value = "MockedNamePropXYBaseDataModule"

        # Assign a static variable READER with ProteinDataReader (to get rid of default Abstract DataReader)
        # Mock Data Reader
        ReaderMock = MagicMock()
        ReaderMock.name.return_value = "MockedReader"
        XYBaseDataModule.READER = ReaderMock

        # Initialize the module with a label_filter
        cls.module = XYBaseDataModule(
            label_filter=1,  # Provide a label_filter
            balance_after_filter=1.0,  # Balance ratio
        )

    def test_filter_labels_valid_index(self) -> None:
        """
        Test the _filter_labels method with a valid label_filter index.
        """
        self.module.label_filter = 1
        row = {
            "features": ["feature1", "feature2"],
            "labels": [0, 3, 1, 2],  # List of labels
        }
        filtered_row = self.module._filter_labels(row)
        expected_labels = [3]  # Only the label at index 1 should be kept

        self.assertEqual(
            filtered_row["labels"],
            expected_labels,
            "The filtered labels do not match the expected labels.",
        )

        row = {
            "features": ["feature1", "feature2"],
            "labels": [True, False, True, True],
        }
        self.assertEqual(
            self.module._filter_labels(row)["labels"],
            [False],
            "The filtered labels for the boolean case do not match the expected labels.",
        )

    def test_filter_labels_no_filter(self) -> None:
        """
        Test the _filter_labels method with no label_filter index.
        """
        # Update the module to have no label filter
        self.module.label_filter = None
        row = {"features": ["feature1", "feature2"], "labels": [False, True]}
        # Handle the case where the index is out of bounds
        with self.assertRaises(
            TypeError, msg="Expected a TypeError when no label filter is provided."
        ):
            self.module._filter_labels(row)

    def test_filter_labels_invalid_index(self) -> None:
        """
        Test the _filter_labels method with an invalid label_filter index.
        """
        # Set an invalid label filter index (e.g., greater than the number of labels)
        self.module.label_filter = 10
        row = {"features": ["feature1", "feature2"], "labels": [False, True]}
        # Handle the case where the index is out of bounds
        with self.assertRaises(
            IndexError,
            msg="Expected an IndexError when the label filter index is out of bounds.",
        ):
            self.module._filter_labels(row)


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Tuple
from unittest.mock import MagicMock, PropertyMock, patch

import pandas as pd

from chebai.preprocessing.datasets.base import _DynamicDataset


class TestDynamicDataset(unittest.TestCase):
    """
    Test case for _DynamicDataset functionality, ensuring correct data splits and integrity
    of train, validation, and test datasets.
    """

    @classmethod
    @patch.multiple(_DynamicDataset, __abstractmethods__=frozenset())
    @patch.object(_DynamicDataset, "base_dir", new_callable=PropertyMock)
    @patch.object(_DynamicDataset, "_name", new_callable=PropertyMock)
    @patch("os.makedirs", return_value=None)
    def setUpClass(
        cls,
        mock_makedirs,
        mock_base_dir_property: PropertyMock,
        mock_name_property: PropertyMock,
    ) -> None:
        """
        Set up a base instance of _DynamicDataset for testing with mocked properties.
        """

        # Mocking properties
        mock_base_dir_property.return_value = "MockedBaseDirPropertyDynamicDataset"
        mock_name_property.return_value = "MockedNamePropertyDynamicDataset"

        # Mock Data Reader
        ReaderMock = MagicMock()
        ReaderMock.name.return_value = "MockedReader"
        _DynamicDataset.READER = ReaderMock

        # Creating an instance of the dataset
        cls.dataset: _DynamicDataset = _DynamicDataset()

        # Dataset with a balanced distribution of labels
        X = [
            [1, 2],
            [3, 4],
            [5, 6],
            [7, 8],
            [9, 10],
            [11, 12],
            [13, 14],
            [15, 16],
            [17, 18],
            [19, 20],
            [21, 22],
            [23, 24],
            [25, 26],
            [27, 28],
            [29, 30],
            [31, 32],
        ]
        y = [
            [False, False],
            [False, True],
            [True, False],
            [True, True],
            [False, False],
            [False, True],
            [True, False],
            [True, True],
            [False, False],
            [False, True],
            [True, False],
            [True, True],
            [False, False],
            [False, True],
            [True, False],
            [True, True],
        ]
        cls.data_df = pd.DataFrame(
            {"ident": [f"id{i + 1}" for i in range(len(X))], "features": X, "labels": y}
        )

    def test_get_test_split_valid(self) -> None:
        """
        Test splitting the dataset into train and test sets and verify balance and non-overlap.
        """
        self.dataset.train_split = 0.5
        # Test size will be 0.25 * 16 = 4
        train_df, test_df = self.dataset.get_test_split(self.data_df, seed=0)

        # Assert the correct number of rows in train and test sets
        self.assertEqual(len(train_df), 12, "Train set should contain 12 samples.")
        self.assertEqual(len(test_df), 4, "Test set should contain 4 samples.")

        # Check positive and negative label counts in train and test sets
        train_pos_count, train_neg_count = self.get_positive_negative_labels_counts(
            train_df
        )
        test_pos_count, test_neg_count = self.get_positive_negative_labels_counts(
            test_df
        )

        # Ensure that the train and test sets have balanced positives and negatives
        self.assertEqual(
            train_pos_count, train_neg_count, "Train set labels should be balanced."
        )
        self.assertEqual(
            test_pos_count, test_neg_count, "Test set labels should be balanced."
        )

        # Assert there is no overlap between train and test sets
        train_idents = set(train_df["ident"])
        test_idents = set(test_df["ident"])
        self.assertEqual(
            len(train_idents.intersection(test_idents)),
            0,
            "Train and test sets should not overlap.",
        )

    def test_get_test_split_missing_labels(self) -> None:
        """
        Test the behavior when the 'labels' column is missing in the dataset.
        """
        df_missing_labels = pd.DataFrame({"ident": ["id1", "id2"]})
        with self.assertRaises(
            KeyError, msg="Expected KeyError when 'labels' column is missing."
        ):
            self.dataset.get_test_split(df_missing_labels)

    def test_get_test_split_seed_consistency(self) -> None:
        """
        Test that splitting the dataset with the same seed produces consistent results.
        """
        train_df1, test_df1 = self.dataset.get_test_split(self.data_df, seed=42)
        train_df2, test_df2 = self.dataset.get_test_split(self.data_df, seed=42)

        pd.testing.assert_frame_equal(
            train_df1,
            train_df2,
            obj="Train sets should be identical for the same seed.",
        )
        pd.testing.assert_frame_equal(
            test_df1, test_df2, obj="Test sets should be identical for the same seed."
        )

    def test_get_train_val_splits_given_test(self) -> None:
        """
        Test splitting the dataset into train and validation sets and verify balance and non-overlap.
        """
        self.dataset.use_inner_cross_validation = False
        self.dataset.train_split = 0.5
        df_train_main, test_df = self.dataset.get_test_split(self.data_df, seed=0)
        train_df, val_df = self.dataset.get_train_val_splits_given_test(
            df_train_main, test_df, seed=42
        )

        # Ensure there is no overlap between train and test sets
        train_idents = set(train_df["ident"])
        test_idents = set(test_df["ident"])
        self.assertEqual(
            len(train_idents.intersection(test_idents)),
            0,
            "Train and test sets should not overlap.",
        )

        # Ensure there is no overlap between validation and test sets
        val_idents = set(val_df["ident"])
        self.assertEqual(
            len(val_idents.intersection(test_idents)),
            0,
            "Validation and test sets should not overlap.",
        )

        # Ensure there is no overlap between train and validation sets
        self.assertEqual(
            len(train_idents.intersection(val_idents)),
            0,
            "Train and validation sets should not overlap.",
        )

        # Check positive and negative label counts in train and validation sets
        train_pos_count, train_neg_count = self.get_positive_negative_labels_counts(
            train_df
        )
        val_pos_count, val_neg_count = self.get_positive_negative_labels_counts(val_df)

        # Ensure that the train and validation sets have balanced positives and negatives
        self.assertEqual(
            train_pos_count, train_neg_count, "Train set labels should be balanced."
        )
        self.assertEqual(
            val_pos_count, val_neg_count, "Validation set labels should be balanced."
        )

    def test_get_train_val_splits_given_test_consistency(self) -> None:
        """
        Test that splitting the dataset into train and validation sets with the same seed produces consistent results.
        """
        test_df = self.data_df.iloc[12:]  # Assume rows 12 onward are for testing
        train_df1, val_df1 = self.dataset.get_train_val_splits_given_test(
            self.data_df, test_df, seed=42
        )
        train_df2, val_df2 = self.dataset.get_train_val_splits_given_test(
            self.data_df, test_df, seed=42
        )

        pd.testing.assert_frame_equal(
            train_df1,
            train_df2,
            obj="Train sets should be identical for the same seed.",
        )
        pd.testing.assert_frame_equal(
            val_df1,
            val_df2,
            obj="Validation sets should be identical for the same seed.",
        )

    def test_get_test_split_stratification(self) -> None:
        """
        Test that the split into train and test sets maintains the stratification of labels.
        """
        self.dataset.train_split = 0.5
        train_df, test_df = self.dataset.get_test_split(self.data_df, seed=0)

        number_of_labels = len(self.data_df["labels"][0])

        # Check the label distribution in the original dataset
        original_pos_count, original_neg_count = (
            self.get_positive_negative_labels_counts(self.data_df)
        )
        total_count = len(self.data_df) * number_of_labels

        # Calculate the expected proportions
        original_pos_proportion = original_pos_count / total_count
        original_neg_proportion = original_neg_count / total_count

        # Check the label distribution in the train set
        train_pos_count, train_neg_count = self.get_positive_negative_labels_counts(
            train_df
        )
        train_total_count = len(train_df) * number_of_labels

        # Calculate the train set proportions
        train_pos_proportion = train_pos_count / train_total_count
        train_neg_proportion = train_neg_count / train_total_count

        # Assert that the proportions are similar to the original dataset
        self.assertAlmostEqual(
            train_pos_proportion,
            original_pos_proportion,
            places=1,
            msg="Train set labels should maintain original positive label proportion.",
        )
        self.assertAlmostEqual(
            train_neg_proportion,
            original_neg_proportion,
            places=1,
            msg="Train set labels should maintain original negative label proportion.",
        )

        # Check the label distribution in the test set
        test_pos_count, test_neg_count = self.get_positive_negative_labels_counts(
            test_df
        )
        test_total_count = len(test_df) * number_of_labels

        # Calculate the test set proportions
        test_pos_proportion = test_pos_count / test_total_count
        test_neg_proportion = test_neg_count / test_total_count

        # Assert that the proportions are similar to the original dataset
        self.assertAlmostEqual(
            test_pos_proportion,
            original_pos_proportion,
            places=1,
            msg="Test set labels should maintain original positive label proportion.",
        )
        self.assertAlmostEqual(
            test_neg_proportion,
            original_neg_proportion,
            places=1,
            msg="Test set labels should maintain original negative label proportion.",
        )

    def test_get_train_val_splits_given_test_stratification(self) -> None:
        """
        Test that the split into train and validation sets maintains the stratification of labels.
        """
        self.dataset.use_inner_cross_validation = False
        self.dataset.train_split = 0.5
        df_train_main, test_df = self.dataset.get_test_split(self.data_df, seed=0)
        train_df, val_df = self.dataset.get_train_val_splits_given_test(
            df_train_main, test_df, seed=42
        )

        number_of_labels = len(self.data_df["labels"][0])

        # Check the label distribution in the original dataset
        original_pos_count, original_neg_count = (
            self.get_positive_negative_labels_counts(self.data_df)
        )
        total_count = len(self.data_df) * number_of_labels

        # Calculate the expected proportions
        original_pos_proportion = original_pos_count / total_count
        original_neg_proportion = original_neg_count / total_count

        # Check the label distribution in the train set
        train_pos_count, train_neg_count = self.get_positive_negative_labels_counts(
            train_df
        )
        train_total_count = len(train_df) * number_of_labels

        # Calculate the train set proportions
        train_pos_proportion = train_pos_count / train_total_count
        train_neg_proportion = train_neg_count / train_total_count

        # Assert that the proportions are similar to the original dataset
        self.assertAlmostEqual(
            train_pos_proportion,
            original_pos_proportion,
            places=1,
            msg="Train set labels should maintain original positive label proportion.",
        )
        self.assertAlmostEqual(
            train_neg_proportion,
            original_neg_proportion,
            places=1,
            msg="Train set labels should maintain original negative label proportion.",
        )

        # Check the label distribution in the validation set
        val_pos_count, val_neg_count = self.get_positive_negative_labels_counts(val_df)
        val_total_count = len(val_df) * number_of_labels

        # Calculate the validation set proportions
        val_pos_proportion = val_pos_count / val_total_count
        val_neg_proportion = val_neg_count / val_total_count

        # Assert that the proportions are similar to the original dataset
        self.assertAlmostEqual(
            val_pos_proportion,
            original_pos_proportion,
            places=1,
            msg="Validation set labels should maintain original positive label proportion.",
        )
        self.assertAlmostEqual(
            val_neg_proportion,
            original_neg_proportion,
            places=1,
            msg="Validation set labels should maintain original negative label proportion.",
        )

    @staticmethod
    def get_positive_negative_labels_counts(df: pd.DataFrame) -> Tuple[int, int]:
        """
        Count the number of True and False values within the labels column.

        Args:
            df (pd.DataFrame): The DataFrame containing the 'labels' column.

        Returns:
            Tuple[int, int]: A tuple containing the counts of True and False values, respectively.
        """
        true_count = sum(sum(label) for label in df["labels"])
        false_count = sum(len(label) - sum(label) for label in df["labels"])
        return true_count, false_count


if __name__ == "__main__":
    unittest.main()



import os
import shutil
from typing import Dict, List, Optional, Tuple, Type

import pandas as pd
import torch
from jsonargparse import CLI

from chebai.preprocessing.datasets.chebi import ChEBIOverXPartial, _ChEBIDataExtractor


class ChebiDataMigration:
    """
    A class to handle migration of ChEBI dataset to a new structure.

    Attributes:
        __MODULE_PATH (str): The path to the module containing ChEBI classes.
        __DATA_ROOT_DIR (str): The root directory for data.
        _chebi_cls (_ChEBIDataExtractor): The ChEBI class instance.
    """

    __MODULE_PATH: str = "chebai.preprocessing.datasets.chebi"
    __DATA_ROOT_DIR: str = "data"

    def __init__(self, datamodule: _ChEBIDataExtractor):
        self._chebi_cls = datamodule

    @classmethod
    def from_args(cls, class_name: str, chebi_version: int, single_class: int = None):
        """
        Class method to create an instance of the class using the provided arguments.

        Args:
            class_name (str): The name of the ChEBI class to be dynamically imported.
            chebi_version (int): The version of ChEBI to be used.
            single_class (int, optional): A specific class identifier if single class extraction is needed.
                Default is None.

        Returns:
            Subclass of _ChEBIDataExtractor
        """
        chebi_cls: _ChEBIDataExtractor = ChebiDataMigration._dynamic_import_chebi_cls(
            class_name, chebi_version, single_class
        )
        return cls(chebi_cls)

    @classmethod
    def _dynamic_import_chebi_cls(
        cls, class_name: str, chebi_version: int, single_class: int
    ) -> _ChEBIDataExtractor:
        """
        Dynamically import the ChEBI class.

        Args:
            class_name (str): The name of the ChEBI class.
            chebi_version (int): The version of the ChEBI dataset.
            single_class (int): The ID of the single class to predict.

        Returns:
            _ChEBIDataExtractor: An instance of the dynamically imported class.
        """
        class_name = class_name.strip()
        module = __import__(cls.__MODULE_PATH, fromlist=[class_name])
        _class = getattr(module, class_name)
        return _class(**{"chebi_version": chebi_version, "single_class": single_class})

    def migrate(self) -> None:
        """
        Start the migration process for the ChEBI dataset.
        """
        os.makedirs(self._chebi_cls.base_dir, exist_ok=True)
        print("Migration started.....")
        old_raw_data_exists = self._migrate_old_raw_data()

        # Either we can combine `.pt` split files to form `data.pt` file
        # self._migrate_old_processed_data()
        # OR
        # we can transform `data.pkl` to `data.pt` file (this seems efficient along with less code)
        if old_raw_data_exists:
            self._chebi_cls.setup_processed()
        else:
            self._migrate_old_processed_data()
        print("Migration completed.....")

    def _migrate_old_raw_data(self) -> bool:
        """
        Migrate old raw data files to the new data folder structure.
        """
        print("-" * 50)
        print("Migrating old raw data....")

        self._copy_file(self._old_raw_dir, self._chebi_cls.raw_dir, "chebi.obo")
        self._copy_file(
            self._old_raw_dir, self._chebi_cls.processed_dir_main, "classes.txt"
        )

        old_splits_file_names_raw = {
            "train": "train.pkl",
            "validation": "validation.pkl",
            "test": "test.pkl",
        }

        data_file_path = os.path.join(self._chebi_cls.processed_dir_main, "data.pkl")
        if os.path.isfile(data_file_path):
            print(f"File {data_file_path} already exists in new data-folder structure")
            return True

        data_df_split_ass_df = self._combine_pkl_splits(
            self._old_raw_dir, old_splits_file_names_raw
        )
        if data_df_split_ass_df is not None:
            data_df = data_df_split_ass_df[0]
            split_ass_df = data_df_split_ass_df[1]
            self._chebi_cls.save_processed(data_df, "data.pkl")
            print(f"File {data_file_path} saved to new data-folder structure")

            split_file = os.path.join(self._chebi_cls.processed_dir_main, "splits.csv")
            split_ass_df.to_csv(split_file)  # overwrites the files with same name
            print(f"File {split_file} saved to new data-folder structure")
            return True
        return False

    def _migrate_old_processed_data(self) -> None:
        """
        Migrate old processed data files to the new data folder structure.
        """
        print("-" * 50)
        print("Migrating old processed data.....")

        data_file_path = os.path.join(self._chebi_cls.processed_dir, "data.pt")
        if os.path.isfile(data_file_path):
            print(f"File {data_file_path} already exists in new data-folder structure")
            return

        old_splits_file_names = {
            "train": "train.pt",
            "validation": "validation.pt",
            "test": "test.pt",
        }

        data_df = self._combine_pt_splits(
            self._old_processed_dir, old_splits_file_names
        )
        if data_df is not None:
            torch.save(data_df, data_file_path)
            print(f"File {data_file_path} saved to new data-folder structure")

    def _combine_pt_splits(
        self, old_dir: str, old_splits_file_names: Dict[str, str]
    ) -> Optional[pd.DataFrame]:
        """
        Combine old `.pt` split files into a single DataFrame.

        Args:
            old_dir (str): The directory containing the old split files.
            old_splits_file_names (Dict[str, str]): A dictionary of split names and file names.

        Returns:
            pd.DataFrame: The combined DataFrame.
        """
        if not self._check_if_old_splits_exists(old_dir, old_splits_file_names):
            print(
                f"Missing at least one of [{', '.join(old_splits_file_names.values())}] in {old_dir}"
            )
            return None

        print("Combining `.pt` splits...")
        df_list: List[pd.DataFrame] = []
        for split, file_name in old_splits_file_names.items():
            file_path = os.path.join(old_dir, file_name)
            file_df = pd.DataFrame(torch.load(file_path, weights_only=False))
            df_list.append(file_df)

        return pd.concat(df_list, ignore_index=True)

    def _combine_pkl_splits(
        self, old_dir: str, old_splits_file_names: Dict[str, str]
    ) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Combine old `.pkl` split files into a single DataFrame and create split assignments.

        Args:
            old_dir (str): The directory containing the old split files.
            old_splits_file_names (Dict[str, str]): A dictionary of split names and file names.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]: The combined DataFrame and split assignments DataFrame.
        """
        if not self._check_if_old_splits_exists(old_dir, old_splits_file_names):
            print(
                f"Missing at least one of [{', '.join(old_splits_file_names.values())}] in {old_dir}"
            )
            return None

        df_list: List[pd.DataFrame] = []
        split_assignment_list: List[pd.DataFrame] = []

        print("Combining `.pkl` splits...")
        for split, file_name in old_splits_file_names.items():
            file_path = os.path.join(old_dir, file_name)
            file_df = pd.read_pickle(file_path)
            df_list.append(file_df)

            # Create split assignment for the current DataFrame
            split_assignment = pd.DataFrame({"id": file_df["id"], "split": split})
            split_assignment_list.append(split_assignment)

        # Concatenate all dataframes and split assignments
        combined_df = pd.concat(df_list, ignore_index=True)
        combined_split_assignment = pd.concat(split_assignment_list, ignore_index=True)

        return combined_df, combined_split_assignment

    @staticmethod
    def _check_if_old_splits_exists(
        old_dir: str, old_splits_file_names: Dict[str, str]
    ) -> bool:
        """
        Check if the old split files exist in the specified directory.

        Args:
            old_dir (str): The directory containing the old split files.
            old_splits_file_names (Dict[str, str]): A dictionary of split names and file names.

        """
        return all(
            os.path.isfile(os.path.join(old_dir, file))
            for file in old_splits_file_names.values()
        )

    @staticmethod
    def _copy_file(old_file_dir: str, new_file_dir: str, file_name: str) -> None:
        """
        Copy a file from the old directory to the new directory.

        Args:
            old_file_dir (str): The directory containing the old file.
            new_file_dir (str): The directory to copy the file to.
            file_name (str): The name of the file to copy.

        Raises:
            FileNotFoundError: If the file does not exist in the old directory.
        """
        os.makedirs(new_file_dir, exist_ok=True)
        new_file_path = os.path.join(new_file_dir, file_name)
        old_file_path = os.path.join(old_file_dir, file_name)

        if os.path.isfile(new_file_path):
            print(
                f"Skipping {old_file_path} (file already exists at new location {new_file_path})"
            )
            return

        if os.path.isfile(old_file_path):
            shutil.copy2(os.path.abspath(old_file_path), os.path.abspath(new_file_path))
            print(f"Copied {old_file_path} to {new_file_path}")
        else:
            print(f"Skipping expected file {old_file_path} (not found)")

    @property
    def _old_base_dir(self) -> str:
        """
        Get the base directory for the old data structure.

        Returns:
            str: The base directory for the old data.
        """
        if isinstance(self._chebi_cls, ChEBIOverXPartial):
            return os.path.join(
                self.__DATA_ROOT_DIR,
                self._chebi_cls._name,
                f"chebi_v{self._chebi_cls.chebi_version}",
                f"partial_{self._chebi_cls.top_class_id}",
            )
        return os.path.join(
            self.__DATA_ROOT_DIR,
            self._chebi_cls._name,
            f"chebi_v{self._chebi_cls.chebi_version}",
        )

    @property
    def _old_processed_dir(self) -> str:
        """
        Get the processed directory for the old data structure.

        Returns:
            str: The processed directory for the old data.
        """
        res = os.path.join(
            self._old_base_dir,
            "processed",
            *self._chebi_cls.identifier,
        )
        if self._chebi_cls.single_class is None:
            return res
        else:
            return os.path.join(res, f"single_{self._chebi_cls.single_class}")

    @property
    def _old_raw_dir(self) -> str:
        """
        Get the raw directory for the old data structure.

        Returns:
            str: The raw directory for the old data.
        """
        return os.path.join(self._old_base_dir, "raw")


class Main:

    def migrate(
        self,
        datamodule: Optional[_ChEBIDataExtractor] = None,
        class_name: Optional[str] = None,
        chebi_version: Optional[int] = None,
        single_class: Optional[int] = None,
    ):
        """
        Migrate ChEBI dataset to new structure and handle splits.

        Args:
            datamodule (Optional[_ChEBIDataExtractor]): The datamodule instance. If not provided, class_name and
            chebi_version are required.
            class_name (Optional[str]): The name of the ChEBI class.
            chebi_version (Optional[int]): The version of the ChEBI dataset.
            single_class (Optional[int]): The ID of the single class to predict.
        """
        if datamodule is not None:
            ChebiDataMigration(datamodule).migrate()
        else:
            ChebiDataMigration.from_args(
                class_name, chebi_version, single_class
            ).migrate()


if __name__ == "__main__":
    CLI(Main)



import csv
import math
import os
import pickle
from typing import List, Literal, Union

import torch

from chebai.loss.bce_weighted import BCEWeighted
from chebai.preprocessing.datasets import XYBaseDataModule
from chebai.preprocessing.datasets.chebi import ChEBIOver100, _ChEBIDataExtractor
from chebai.preprocessing.datasets.pubchem import LabeledUnlabeledMixed


class ImplicationLoss(torch.nn.Module):
    """
    Implication Loss module.

    Args:
        data_extractor _ChEBIDataExtractor: Data extractor for labels.
        base_loss (torch.nn.Module, optional): Base loss function. Defaults to None.
        fuzzy_implication (Literal["product", "lukasiewicz", "xu19"], optional): T-norm type. Defaults to "product".
        impl_loss_weight (float, optional): Weight of implication loss relative to base loss. Defaults to 0.1.
        pos_scalar (int, optional): Positive scalar exponent. Defaults to 1.
        pos_epsilon (float, optional): Epsilon value for numerical stability. Defaults to 0.01.
        multiply_by_softmax (bool, optional): Whether to multiply by softmax. Defaults to False.
        use_sigmoidal_implication (bool, optional): Whether to use the sigmoidal fuzzy implication based on the
        specified fuzzy_implication (as defined by van Krieken et al., 2022: Analyzing Differentiable Fuzzy Logic
        Operators). Defaults to False.
        weight_epoch_dependent (Union[bool, tuple[int, int]], optional): Whether to weight the implication loss
        depending on the current epoch with the sigmoid function sigmoid((epoch-c)/s). If True, c=50 and s=10,
        otherwise, a tuple of integers (c,s) can be supplied. Defaults to False.
        start_at_epoch (int, optional): Epoch at which to start applying the loss. Defaults to 0.
        violations_per_cls_aggregator (Literal["sum", "max"], optional): How to aggregate violations for each class.
        If a class is involved in several implications / disjointnesses, the loss value for this class will be
        aggregated with this method. Defaults to "sum".
    """

    def __init__(
        self,
        data_extractor: XYBaseDataModule,
        base_loss: torch.nn.Module = None,
        fuzzy_implication: Literal[
            "reichenbach",
            "rc",
            "lukasiewicz",
            "lk",
            "xu19",
            "kleene_dienes",
            "kd",
            "goedel",
            "g",
            "reverse-goedel",
            "rg",
            "binary",
            "b",
        ] = "reichenbach",
        impl_loss_weight: float = 0.1,
        pos_scalar: Union[int, float] = 1,
        pos_epsilon: float = 0.01,
        multiply_by_softmax: bool = False,
        use_sigmoidal_implication: bool = False,
        weight_epoch_dependent: Union[bool | tuple[int, int]] = False,
        start_at_epoch: int = 0,
        violations_per_cls_aggregator: Literal[
            "sum", "max", "mean", "log-sum", "log-max", "log-mean"
        ] = "sum",
        multiply_with_base_loss: bool = True,
        no_grads: bool = False,
    ):
        super().__init__()
        # automatically choose labeled subset for implication filter in case of mixed dataset
        if isinstance(data_extractor, LabeledUnlabeledMixed):
            data_extractor = data_extractor.labeled
        assert isinstance(data_extractor, _ChEBIDataExtractor)
        self.data_extractor = data_extractor
        # propagate data_extractor to base loss
        if isinstance(base_loss, BCEWeighted):
            base_loss.data_extractor = self.data_extractor
            base_loss.reduction = (
                "none"  # needed to multiply fuzzy loss with base loss for each sample
            )
        self.base_loss = base_loss
        self.implication_cache_file = f"implications_{self.data_extractor.name}.cache"
        self.label_names = _load_label_names(
            os.path.join(data_extractor.processed_dir_main, "classes.txt")
        )
        self.hierarchy = self._load_implications(
            os.path.join(data_extractor.raw_dir, "chebi.obo")
        )
        implication_filter_dense = _build_dense_filter(
            _build_implication_filter(self.label_names, self.hierarchy),
            len(self.label_names),
        )
        self.implication_filter_l = implication_filter_dense
        self.implication_filter_r = self.implication_filter_l.transpose(0, 1)
        self.fuzzy_implication = fuzzy_implication
        self.impl_weight = impl_loss_weight
        self.pos_scalar = pos_scalar
        self.eps = pos_epsilon
        self.multiply_by_softmax = multiply_by_softmax
        self.use_sigmoidal_implication = use_sigmoidal_implication
        self.weight_epoch_dependent = weight_epoch_dependent
        self.start_at_epoch = start_at_epoch
        self.violations_per_cls_aggregator = violations_per_cls_aggregator
        self.multiply_with_base_loss = multiply_with_base_loss
        self.no_grads = no_grads

    def _calculate_unaggregated_fuzzy_loss(
        self,
        pred,
        target: torch.Tensor,
        weight,
        filter_l,
        filter_r,
        mode="impl",
        **kwargs,
    ):
        # for each batch, get all pairwise losses: [a1, a2, a3] -> [[a1*a1, a1*a2, a1*a3],[a2*a1,...],[a3*a1,...]]
        preds_expanded1 = pred.unsqueeze(1).expand(-1, pred.shape[1], -1)
        preds_expanded2 = pred.unsqueeze(2).expand(-1, -1, pred.shape[1])
        # filter by implication relations and labels

        label_filter = target.unsqueeze(2).expand(-1, -1, pred.shape[1])
        filter_l = filter_l.to(pred.device).unsqueeze(0).expand(pred.shape[0], -1, -1)
        filter_r = filter_r.to(pred.device).unsqueeze(0).expand(pred.shape[0], -1, -1)
        if mode == "impl":
            all_implications = self._calculate_implication_loss(
                preds_expanded2, preds_expanded1
            )
        else:
            all_implications = self._calculate_implication_loss(
                preds_expanded2, 1 - preds_expanded1
            )
        loss_impl_l = all_implications * filter_l * (1 - label_filter)
        if mode == "impl":
            loss_impl_r = all_implications.transpose(1, 2) * filter_r * label_filter
            loss_impl_sum = loss_impl_l + loss_impl_r
        else:
            loss_impl_sum = loss_impl_l

        if self.violations_per_cls_aggregator.startswith("log-"):
            loss_impl_sum = -torch.log(1 - loss_impl_sum)
            violations_per_cls_aggregator = self.violations_per_cls_aggregator[4:]
        else:
            violations_per_cls_aggregator = self.violations_per_cls_aggregator
        if violations_per_cls_aggregator == "sum":
            loss_by_cls = loss_impl_sum.sum(dim=-1)
        elif violations_per_cls_aggregator == "max":
            loss_by_cls = loss_impl_sum.max(dim=-1).values
        elif violations_per_cls_aggregator == "mean":
            loss_by_cls = loss_impl_sum.mean(dim=-1)
        else:
            raise NotImplementedError(
                f"Unknown violations_per_cls_aggregator {self.violations_per_cls_aggregator}"
            )

        unweighted_mean = loss_by_cls.mean()
        implication_loss_weighted = loss_by_cls
        if "current_epoch" in kwargs and self.weight_epoch_dependent:
            sigmoid_center = (
                self.weight_epoch_dependent[0]
                if isinstance(self.weight_epoch_dependent, tuple)
                else 50
            )
            sigmoid_spread = (
                self.weight_epoch_dependent[1]
                if isinstance(self.weight_epoch_dependent, tuple)
                else 10
            )
            # sigmoid function centered around epoch 50
            implication_loss_weighted = implication_loss_weighted / (
                1
                + math.exp(-(kwargs["current_epoch"] - sigmoid_center) / sigmoid_spread)
            )
        implication_loss_weighted *= weight
        weighted_mean = implication_loss_weighted.mean()

        return implication_loss_weighted, unweighted_mean, weighted_mean

    def _calculate_unaggregated_base_loss(self, input, target, **kwargs):
        nnl = kwargs.pop("non_null_labels", None)
        labeled_input = input[nnl] if nnl else input

        if target is not None and self.base_loss is not None:
            return self.base_loss(labeled_input, target.float())
        else:
            return torch.zeros(input.shape, device=input.device)

    def forward(self, input: torch.Tensor, target: torch.Tensor, **kwargs) -> tuple:
        """
        Forward pass of the implication loss module.

        Args:
            input (torch.Tensor): Input tensor.
            target (torch.Tensor): Target tensor.
            **kwargs: Additional arguments.

        Returns:
            tuple: Tuple containing total loss, base loss, and implication loss.
        """
        base_loss = self._calculate_unaggregated_base_loss(input, target, **kwargs)
        loss_components = {"base_loss": base_loss.mean()}

        if "current_epoch" in kwargs and self.start_at_epoch > kwargs["current_epoch"]:
            return base_loss.mean(), loss_components

        pred = torch.sigmoid(input)
        fuzzy_loss, unweighted_fuzzy_mean, weighted_fuzzy_mean = (
            self._calculate_unaggregated_fuzzy_loss(
                pred,
                target,
                self.impl_weight,
                self.implication_filter_l,
                self.implication_filter_r,
                **kwargs,
            )
        )
        if self.no_grads:
            fuzzy_loss = fuzzy_loss.detach()
        loss_components["unweighted_fuzzy_loss"] = unweighted_fuzzy_mean
        loss_components["weighted_fuzzy_loss"] = weighted_fuzzy_mean
        if self.base_loss is None or target is None:
            total_loss = self.impl_weight * fuzzy_loss
        elif self.multiply_with_base_loss:
            total_loss = base_loss * (1 + self.impl_weight * fuzzy_loss)
        else:
            total_loss = base_loss + self.impl_weight * fuzzy_loss
        return total_loss.mean(), loss_components

    def _calculate_implication_loss(
        self, l: torch.Tensor, r: torch.Tensor
    ) -> torch.Tensor:
        """
        Calculate implication loss based on T-norm and other parameters.

        Args:
            l (torch.Tensor): Left part of implication.
            r (torch.Tensor): Right part of implication.

        Returns:
            torch.Tensor: Calculated implication loss.
        """
        assert not l.isnan().any(), (
            f"l contains NaN values - l.shape: {l.shape}, l.isnan().sum(): {l.isnan().sum()}, "
            f"l: {l}"
        )
        assert not r.isnan().any(), (
            f"r contains NaN values - r.shape: {r.shape}, r.isnan().sum(): {r.isnan().sum()}, "
            f"r: {r}"
        )
        if self.pos_scalar != 1:
            l = (
                torch.pow(l + self.eps, 1 / self.pos_scalar)
                - math.pow(self.eps, 1 / self.pos_scalar)
            ) / (
                math.pow(1 + self.eps, 1 / self.pos_scalar)
                - math.pow(self.eps, 1 / self.pos_scalar)
            )
            one_min_r = (
                torch.pow(1 - r + self.eps, 1 / self.pos_scalar)
                - math.pow(self.eps, 1 / self.pos_scalar)
            ) / (
                math.pow(1 + self.eps, 1 / self.pos_scalar)
                - math.pow(self.eps, 1 / self.pos_scalar)
            )
        else:
            one_min_r = 1 - r
        # for each implication I, calculate 1 - I(l, 1-one_min_r)
        # for S-implications, this is equivalent to the t-norm
        if self.fuzzy_implication in ["reichenbach", "rc"]:
            individual_loss = l * one_min_r
        # xu19 (from Xu et al., 2019: Semantic loss) is not a fuzzy implication, but behaves similar to the Reichenbach
        # implication
        elif self.fuzzy_implication == "xu19":
            individual_loss = -torch.log(1 - l * one_min_r)
        elif self.fuzzy_implication in ["lukasiewicz", "lk"]:
            individual_loss = torch.relu(l + one_min_r - 1)
        elif self.fuzzy_implication in ["kleene_dienes", "kd"]:
            individual_loss = torch.min(l, 1 - r)
        elif self.fuzzy_implication in ["goedel", "g"]:
            individual_loss = torch.where(l <= r, 0, one_min_r)
        elif self.fuzzy_implication in ["reverse-goedel", "rg"]:
            individual_loss = torch.where(l <= r, 0, l)
        elif self.fuzzy_implication in ["binary", "b"]:
            individual_loss = torch.where(l <= r, 0, 1).to(dtype=l.dtype)
        else:
            raise NotImplementedError(
                f"Unknown fuzzy implication {self.fuzzy_implication}"
            )

        if self.use_sigmoidal_implication:
            # formula by van Krieken, 2022, applied to fuzzy implication with default parameters: b_0 = 0.5, s = 9
            # parts that only depend on b_0 and s are pre-calculated
            implication = 1 - individual_loss
            sigmoidal_implication = 0.01123379 * (
                91.0171 * torch.sigmoid(9 * (implication - 0.5)) - 1
            )
            individual_loss = 1 - sigmoidal_implication

        if self.multiply_by_softmax:
            individual_loss = individual_loss * individual_loss.softmax(dim=-1)

        return individual_loss

    def _load_implications(self, path_to_chebi: str) -> dict:
        """
        Load class hierarchy implications.

        Args:
            path_to_chebi (str): Path to the ChEBI ontology file.

        Returns:
            dict: Loaded hierarchy of implications.
        """
        if os.path.isfile(self.implication_cache_file):
            with open(self.implication_cache_file, "rb") as fin:
                hierarchy = pickle.load(fin)
        else:
            hierarchy = self.data_extractor.extract_class_hierarchy(path_to_chebi)
            with open(self.implication_cache_file, "wb") as fout:
                pickle.dump(hierarchy, fout)
        return hierarchy


class DisjointLoss(ImplicationLoss):
    """
    Disjoint Loss module, extending ImplicationLoss.

    Args:
        path_to_disjointness (str): Path to the disjointness data file (a csv file containing pairs of disjoint classes)
        data_extractor (Union[_ChEBIDataExtractor, LabeledUnlabeledMixed]): Data extractor for labels.
        base_loss (torch.nn.Module, optional): Base loss function. Defaults to None.
        disjoint_loss_weight (float, optional): Weight of disjointness loss. Defaults to 100.
        **kwargs: Additional arguments.
    """

    def __init__(
        self,
        path_to_disjointness: str,
        data_extractor: Union[_ChEBIDataExtractor, LabeledUnlabeledMixed],
        base_loss: torch.nn.Module = None,
        disjoint_loss_weight: float = 100,
        **kwargs,
    ):
        super().__init__(data_extractor, base_loss, **kwargs)
        self.disjoint_filter_l, self.disjoint_filter_r = _build_disjointness_filter(
            path_to_disjointness, self.label_names, self.hierarchy
        )
        self.disjoint_weight = disjoint_loss_weight

    def forward(self, input: torch.Tensor, target: torch.Tensor, **kwargs) -> tuple:
        """
        Forward pass of the disjoint loss module.

        Args:
            input (torch.Tensor): Input tensor.
            target (torch.Tensor): Target tensor.
            **kwargs: Additional arguments.

        Returns:
            tuple: Tuple containing total loss, base loss, implication loss, and disjointness loss.
        """
        base_loss = self._calculate_unaggregated_base_loss(input, target, **kwargs)
        loss_components = {"base_loss": base_loss.mean()}

        if "current_epoch" in kwargs and self.start_at_epoch > kwargs["current_epoch"]:
            return base_loss.mean(), loss_components

        pred = torch.sigmoid(input)
        impl_loss, unweighted_impl_mean, weighted_impl_mean = (
            self._calculate_unaggregated_fuzzy_loss(
                pred,
                target,
                self.impl_weight,
                self.implication_filter_l,
                self.implication_filter_r,
                **kwargs,
            )
        )
        if self.no_grads:
            impl_loss = impl_loss.detach()
        loss_components["unweighted_implication_loss"] = unweighted_impl_mean
        loss_components["weighted_implication_loss"] = weighted_impl_mean

        disj_loss, unweighted_disj_mean, weighted_disj_mean = (
            self._calculate_unaggregated_fuzzy_loss(
                pred,
                target,
                self.disjoint_weight,
                self.disjoint_filter_l,
                self.disjoint_filter_r,
                mode="disj",
                **kwargs,
            )
        )
        if self.no_grads:
            disj_loss = disj_loss.detach()
        loss_components["unweighted_disjointness_loss"] = unweighted_disj_mean
        loss_components["weighted_disjointness_loss"] = weighted_disj_mean

        if self.base_loss is None or target is None:
            total_loss = self.impl_weight * impl_loss + self.disjoint_weight * disj_loss
        elif self.multiply_with_base_loss:
            total_loss = base_loss * (
                1 + self.impl_weight * impl_loss + self.disjoint_weight * disj_loss
            )
        else:
            total_loss = (
                base_loss
                + self.impl_weight * impl_loss
                + self.disjoint_weight * disj_loss
            )
        return total_loss.mean(), loss_components


def _load_label_names(path_to_label_names: str) -> List:
    """
    Load label names from a file.

    Args:
        path_to_label_names (str): Path to the label names file.

    Returns:
        list: List of label names.
    """
    with open(path_to_label_names) as fin:
        label_names = [int(line.strip()) for line in fin]
    return label_names


def _build_implication_filter(label_names: List, hierarchy: dict) -> torch.Tensor:
    """
    Build implication filter based on label names and hierarchy. Results in list of pairs (A,B) for each implication
    A->B (including indirect implications).

    Args:
        label_names (list): List of label names.
        hierarchy (dict): Hierarchy of implications.

    Returns:
        torch.Tensor: Tensor representing implication filter.
    """
    return torch.tensor(
        [
            (i1, i2)
            for i1, l1 in enumerate(label_names)
            for i2, l2 in enumerate(label_names)
            if l2 in hierarchy.pred[l1]
        ]
    )


def _build_dense_filter(sparse_filter: torch.Tensor, n_labels: int) -> torch.Tensor:
    res = torch.zeros((n_labels, n_labels), dtype=torch.bool)
    for l, r in sparse_filter:
        res[l, r] = True
    return res


def _build_disjointness_filter(
    path_to_disjointness: str, label_names: List, hierarchy: dict
) -> tuple:
    """
    Build disjointness filter based on disjointness data and hierarchy.

    Args:
        path_to_disjointness (str): Path to the disjointness data file.
        label_names (list): List of label names.
        hierarchy (dict): Hierarchy of implications.

    Returns:
        tuple: Tuple containing tensors representing disjointness filter.
    """
    disjoints = set()
    label_dict = dict(map(reversed, enumerate(label_names)))

    with open(path_to_disjointness, "rt") as fin:
        reader = csv.reader(fin)
        for l1_raw, r1_raw in reader:
            l1 = int(l1_raw)
            r1 = int(r1_raw)
            if l1 == 36233 and r1 == 63353:
                # ignore disaccharide-disaccharide derivative disjointness axiom
                continue
            disjoints.update(
                {
                    (label_dict[l2], label_dict[r2])
                    for r2 in list(hierarchy.succ[r1]) + [r1]
                    if r2 in label_names
                    for l2 in list(hierarchy.succ[l1]) + [l1]
                    if l2 in label_names
                }
            )

    dis_filter = torch.tensor(list(disjoints))
    dense = _build_dense_filter(dis_filter, len(label_names))
    dense_r = dense.transpose(0, 1)
    return dense, dense_r


if __name__ == "__main__":
    loss = DisjointLoss(
        os.path.join("data", "disjoint.csv"),
        ChEBIOver100(chebi_version=231),
        base_loss=BCEWeighted(),
        impl_loss_weight=1,
        disjoint_loss_weight=1,
    )
    random_preds = torch.randn(10, 997)
    random_labels = torch.randint(0, 2, (10, 997))
    for agg in ["sum", "max", "mean", "log-mean"]:
        loss.violations_per_cls_aggregator = agg
        l = loss(random_preds, random_labels)
        print(f"Loss with {agg} aggregation for random input:", l)

    # simplified example for ontology with 4 classes, A -> B, B -> C, D -> C, B and D disjoint
    loss.implication_filter_l = torch.tensor(
        [[0, 1, 1, 0], [0, 0, 1, 0], [0, 0, 0, 0], [0, 0, 1, 0]]
    )
    loss.implication_filter_r = loss.implication_filter_l.transpose(0, 1)
    loss.disjoint_filter_l = torch.tensor(
        [[0, 0, 0, 0], [0, 0, 0, 1], [0, 0, 0, 0], [0, 1, 0, 0]]
    )
    loss.disjoint_filter_r = loss.disjoint_filter_l.transpose(0, 1)
    # expected result: first sample: moderately high loss for B disj D, otherwise low, second sample: high loss for A -> B (applied to A), otherwise low
    preds = torch.tensor([[0.1, 0.3, 0.7, 0.4], [0.5, 0.2, 0.9, 0.1]])
    labels = [[0, 1, 1, 0], [0, 0, 1, 1]]
    for agg in ["sum", "max", "mean", "log-mean"]:
        loss.violations_per_cls_aggregator = agg
        l = loss(preds, torch.tensor(labels))
        print(f"Loss with {agg} aggregation for simple input:", l)



import unittest
from unittest.mock import MagicMock, PropertyMock, mock_open, patch

import networkx as nx
import pandas as pd

from chebai.preprocessing.datasets.chebi import _ChEBIDataExtractor
from tests.unit.mock_data.ontology_mock_data import ChebiMockOntology


class TestChEBIDataExtractor(unittest.TestCase):

    @classmethod
    @patch.multiple(_ChEBIDataExtractor, __abstractmethods__=frozenset())
    @patch.object(_ChEBIDataExtractor, "base_dir", new_callable=PropertyMock)
    @patch.object(_ChEBIDataExtractor, "_name", new_callable=PropertyMock)
    @patch("os.makedirs", return_value=None)
    def setUpClass(
        cls,
        mock_makedirs,
        mock_name_property: PropertyMock,
        mock_base_dir_property: PropertyMock,
    ) -> None:
        """
        Set up a base instance of _ChEBIDataExtractor for testing with mocked properties.
        """
        # Mocking properties
        mock_base_dir_property.return_value = "MockedBaseDirPropertyChebiDataExtractor"
        mock_name_property.return_value = "MockedNamePropertyChebiDataExtractor"

        # Mock Data Reader
        ReaderMock = MagicMock()
        ReaderMock.name.return_value = "MockedReader"
        _ChEBIDataExtractor.READER = ReaderMock

        # Create an instance of the dataset
        cls.extractor: _ChEBIDataExtractor = _ChEBIDataExtractor(
            chebi_version=231, chebi_version_train=200
        )

        # Mock instance for _chebi_version_train_obj
        mock_train_obj = MagicMock()
        mock_train_obj.processed_dir_main = "/mock/path/to/train"
        cls.extractor._chebi_version_train_obj = mock_train_obj

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=ChebiMockOntology.get_raw_data(),
    )
    def test_extract_class_hierarchy(self, mock_open: mock_open) -> None:
        """
        Test the extraction of class hierarchy and validate the structure of the resulting graph.
        """
        # Mock the output of fastobo.loads
        graph = self.extractor._extract_class_hierarchy("fake_path")

        # Validate the graph structure
        self.assertIsInstance(
            graph, nx.DiGraph, "The result should be a directed graph."
        )

        # Check nodes
        actual_nodes = set(graph.nodes)
        self.assertEqual(
            set(ChebiMockOntology.get_nodes()),
            actual_nodes,
            "The graph nodes do not match the expected nodes.",
        )

        # Check edges
        actual_edges = set(graph.edges)
        self.assertEqual(
            ChebiMockOntology.get_edges_of_transitive_closure_graph(),
            actual_edges,
            "The graph edges do not match the expected edges.",
        )

        # Check number of nodes and edges
        self.assertEqual(
            ChebiMockOntology.get_number_of_nodes(),
            len(actual_nodes),
            "The number of nodes should match the actual number of nodes in the graph.",
        )

        self.assertEqual(
            ChebiMockOntology.get_number_of_transitive_edges(),
            len(actual_edges),
            "The number of transitive edges should match the actual number of transitive edges in the graph.",
        )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=ChebiMockOntology.get_raw_data(),
    )
    @patch.object(
        _ChEBIDataExtractor,
        "select_classes",
        return_value=ChebiMockOntology.get_nodes(),
    )
    def test_graph_to_raw_dataset(
        self, mock_select_classes: PropertyMock, mock_open: mock_open
    ) -> None:
        """
        Test conversion of a graph to a raw dataset and compare it with the expected DataFrame.
        """
        graph = self.extractor._extract_class_hierarchy("fake_path")
        data_df = self.extractor._graph_to_raw_dataset(graph)

        pd.testing.assert_frame_equal(
            data_df,
            ChebiMockOntology.get_data_in_dataframe(),
            obj="The DataFrame should match the expected structure.",
        )

    @patch(
        "builtins.open", new_callable=mock_open, read_data=b"Mocktestdata"
    )  # Mocking open as a binary file
    @patch("pandas.read_pickle")
    def test_load_dict(
        self, mock_read_pickle: PropertyMock, mock_open: mock_open
    ) -> None:
        """
        Test loading data from a pickled file and verify the generator output.
        """
        # Mock the DataFrame returned by read_pickle
        mock_df = pd.DataFrame(
            {
                "id": [12345, 67890, 11111, 54321],  # Corrected ID
                "name": ["A", "B", "C", "D"],
                "SMILES": ["C1CCCCC1", "O=C=O", "C1CC=CC1", "C[Mg+]"],
                12345: [True, False, False, True],
                67890: [False, True, True, False],
                11111: [True, False, True, False],
            }
        )
        mock_read_pickle.return_value = mock_df

        generator = self.extractor._load_dict("data/tests")
        result = list(generator)

        # Convert NumPy arrays to lists for comparison
        for item in result:
            item["labels"] = list(item["labels"])

        # Expected output for comparison
        expected_result = [
            {"features": "C1CCCCC1", "labels": [True, False, True], "ident": 12345},
            {"features": "O=C=O", "labels": [False, True, False], "ident": 67890},
            {"features": "C1CC=CC1", "labels": [False, True, True], "ident": 11111},
            {"features": "C[Mg+]", "labels": [True, False, False], "ident": 54321},
        ]

        # Assert if the result matches the expected output
        self.assertEqual(
            result,
            expected_result,
            "The loaded dictionary should match the expected structure.",
        )

    @patch("builtins.open", new_callable=mock_open)
    @patch.object(_ChEBIDataExtractor, "processed_dir_main", new_callable=PropertyMock)
    def test_setup_pruned_test_set(
        self,
        mock_processed_dir_main: PropertyMock,
        mock_open_file: mock_open,
    ) -> None:
        """
        Test the pruning of the test set to match classes in the training set.
        """
        # Mock the content for the two open calls (original classes and new classes)
        mock_orig_classes = "12345\n67890\n88888\n54321\n77777\n"
        mock_new_classes = "12345\n67890\n99999\n77777\n"

        # Use side_effect to simulate the two different file reads
        mock_open_file.side_effect = [
            mock_open(
                read_data=mock_orig_classes
            ).return_value,  # First open() for orig_classes
            mock_open(
                read_data=mock_new_classes
            ).return_value,  # Second open() for new_classes
        ]

        # Mock the attributes used in the method
        mock_processed_dir_main.return_value = "/mock/path/to/current"

        # Mock DataFrame to simulate the test dataset
        mock_df = pd.DataFrame(
            {
                "labels": [
                    [
                        True,
                        False,
                        True,
                        False,
                        True,
                    ],  # First test instance labels (match orig_classes)
                    [False, True, False, True, False],
                ]  # Second test instance labels
            }
        )

        # Call the method under test
        pruned_df = self.extractor._setup_pruned_test_set(mock_df)

        # Expected DataFrame labels after pruning (only "12345", "67890", "77777", and "99999" remain)
        expected_labels = [[True, False, False, True], [False, True, False, False]]

        # Check if the pruned DataFrame still has the same number of rows
        self.assertEqual(
            len(pruned_df),
            len(mock_df),
            "The pruned DataFrame should have the same number of rows.",
        )

        # Check that the labels are correctly pruned
        for i in range(len(pruned_df)):
            self.assertEqual(
                pruned_df.iloc[i]["labels"],
                expected_labels[i],
                f"Row {i}'s labels should be pruned correctly.",
            )


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Any, Dict

import fastobo
from fastobo.term import TermFrame

from chebai.preprocessing.datasets.chebi import term_callback
from tests.unit.mock_data.ontology_mock_data import ChebiMockOntology


class TestChebiTermCallback(unittest.TestCase):
    """
    Unit tests for the `term_callback` function used in processing ChEBI ontology terms.
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up the test class by loading ChEBI term data and storing it in a dictionary
        where keys are the term IDs and values are TermFrame instances.
        """
        cls.callback_input_data: Dict[int, TermFrame] = {
            int(term_doc.id.local): term_doc
            for term_doc in fastobo.loads(ChebiMockOntology.get_raw_data())
            if term_doc and ":" in str(term_doc.id)
        }

    def test_process_valid_terms(self) -> None:
        """
        Test that `term_callback` correctly processes valid ChEBI terms.
        """

        expected_result: Dict[str, Any] = {
            "id": 12345,
            "parents": [54321, 67890],
            "has_part": set(),
            "name": "Compound A",
            "smiles": "C1=CC=CC=C1",
        }

        actual_dict: Dict[str, Any] = term_callback(
            self.callback_input_data.get(expected_result["id"])
        )
        self.assertEqual(
            expected_result,
            actual_dict,
            msg="term_callback should correctly extract information from valid ChEBI terms.",
        )

    def test_skip_obsolete_terms(self) -> None:
        """
        Test that `term_callback` correctly skips obsolete ChEBI terms.
        """
        term_callback_output = []
        for ident in ChebiMockOntology.get_obsolete_nodes_ids():
            raw_term = self.callback_input_data.get(ident)
            term_dict = term_callback(raw_term)
            if term_dict:
                term_callback_output.append(term_dict)

        self.assertEqual(
            term_callback_output,
            [],
            msg="The term_callback function should skip obsolete terms and return an empty list.",
        )


if __name__ == "__main__":
    unittest.main()



import csv
import gzip
import os
import shutil
import zipfile
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Dict, Generator, List, Optional
from urllib import request

import numpy as np
import torch
from rdkit import Chem
from sklearn.model_selection import GroupShuffleSplit, train_test_split

from chebai.preprocessing import reader as dr
from chebai.preprocessing.datasets.base import XYBaseDataModule


class Tox21MolNet(XYBaseDataModule):
    """Data module for Tox21MolNet dataset."""

    HEADERS = [
        "NR-AR",
        "NR-AR-LBD",
        "NR-AhR",
        "NR-Aromatase",
        "NR-ER",
        "NR-ER-LBD",
        "NR-PPAR-gamma",
        "SR-ARE",
        "SR-ATAD5",
        "SR-HSE",
        "SR-MMP",
        "SR-p53",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "Tox21MN"

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["tox21.csv"]

    @property
    def processed_file_names(self) -> List[str]:
        """Returns a list of processed file names."""
        return ["test.pt", "train.pt", "validation.pt"]

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with NamedTemporaryFile("rb") as gout:
            request.urlretrieve(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/tox21.csv.gz",
                gout.name,
            )
            with gzip.open(gout.name) as gfile:
                with open(os.path.join(self.raw_dir, "tox21.csv"), "wt") as fout:
                    fout.write(gfile.read().decode())

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = self._load_data_from_file(os.path.join(self.raw_dir, f"tox21.csv"))
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            split_size = int(len(set(groups)) * self.train_split)
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(len(set(split_groups)) * self.train_split), n_splits=1
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d
                for d in (data[temp_split_index[i]] for i in test_split_index)
                if d["original"]
            ]
            validation_split = [
                d
                for d in (data[temp_split_index[i]] for i in validation_split_index)
                if d["original"]
            ]
        else:
            train_split, test_split = train_test_split(
                data, train_size=self.train_split, shuffle=True
            )
            test_split, validation_split = train_test_split(
                test_split, train_size=0.5, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if self._setup_data_flag != 1:
            return

        self._setup_data_flag += 1
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._set_processed_data_props()

    def _load_data_from_file(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                smiles = row["smiles"]
                labels = [
                    bool(int(l)) if l else None for l in (row[k] for k in self.HEADERS)
                ]
                yield dict(features=smiles, labels=labels, ident=row["mol_id"])


class Tox21Challenge(XYBaseDataModule):
    """Data module for Tox21Challenge dataset."""

    HEADERS = [
        "NR-AR",
        "NR-AR-LBD",
        "NR-AhR",
        "NR-Aromatase",
        "NR-ER",
        "NR-ER-LBD",
        "NR-PPAR-gamma",
        "SR-ARE",
        "SR-ATAD5",
        "SR-HSE",
        "SR-MMP",
        "SR-p53",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "Tox21Chal"

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return [
            "train.sdf",
            "validation.sdf",
            "validation.smiles",
            "test.smiles",
            "test_results.txt",
        ]

    @property
    def processed_file_names(self) -> List[str]:
        """Returns a list of processed file names."""
        return ["test.pt", "train.pt", "validation.pt"]

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        self._retrieve_file(
            "https://tripod.nih.gov/tox21/challenge/download?id=tox21_10k_data_allsdf&sec=",
            "train.sdf",
            compression="zip",
        )
        self._retrieve_file(
            "https://tripod.nih.gov/tox21/challenge/download?id=tox21_10k_challenge_testsdf&sec=",
            "validation.sdf",
            compression="zip",
        )
        self._retrieve_file(
            "https://tripod.nih.gov/tox21/challenge/download?id=tox21_10k_challenge_scoresmiles&sec=",
            "test.smiles",
        )
        self._retrieve_file(
            "https://tripod.nih.gov/tox21/challenge/download?id=tox21_10k_challenge_scoretxt&sec=",
            "test_results.txt",
        )

    def _retrieve_file(
        self, url: str, target_file: str, compression: Optional[str] = None
    ) -> None:
        """Retrieves a file from a URL and saves it locally.

        Args:
            url (str): The URL to download the file from.
            target_file (str): The name of the target file.
            compression (str, optional): Compression type. Defaults to None.
        """
        target_path = os.path.join(self.raw_dir, target_file)
        if not os.path.isfile(target_path):
            with NamedTemporaryFile("rb") as gout:
                if compression is None:
                    download_path = target_path
                else:
                    download_path = gout.name
                request.urlretrieve(
                    url,
                    download_path,
                )
                if compression == "zip":
                    td = TemporaryDirectory()
                    with zipfile.ZipFile(download_path, "r") as zip_ref:
                        zip_ref.extractall(td.name)
                        files_in_zip = os.listdir(td.name)
                        f = files_in_zip[0]
                        assert len(files_in_zip) == 1
                        shutil.move(os.path.join(td.name, f), target_path)

    def _load_data_from_file(self, path: str) -> List[Dict]:
        """Loads data from an SDF file.

        Args:
            path (str): Path to the SDF file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        sdf = Chem.SDMolSupplier(path)
        data = []
        for mol in sdf:
            if mol is not None:
                d = dict(
                    labels=[
                        int(mol.GetProp(h)) if h in mol.GetPropNames() else None
                        for h in self.HEADERS
                    ],
                    ident=[
                        mol.GetProp(k)
                        for k in ("DSSTox_CID", "Compound ID")
                        if k in mol.GetPropNames()
                    ][0],
                    features=Chem.MolToSmiles(mol),
                )
                data.append(self.reader.to_data(d))
        return data

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        for k in ("train", "validation"):
            d = self._load_data_from_file(os.path.join(self.raw_dir, f"{k}.sdf"))
            torch.save(d, os.path.join(self.processed_dir, f"{k}.pt"))

        with open(os.path.join(self.raw_dir, f"test.smiles")) as fin:
            next(fin)
            test_smiles = dict(reversed(row.strip().split("\t")) for row in fin)
        with open(os.path.join(self.raw_dir, f"test_results.txt")) as fin:
            headers = next(fin).strip().split("\t")
            test_results = {
                k["Sample ID"]: [
                    int(k[h]) if k[h] != "x" else None for h in self.HEADERS
                ]
                for k in (
                    dict(zip(headers, row.strip().split("\t"))) for row in fin if row
                )
            }
        test_data = [
            self.reader.to_data(
                dict(features=test_smiles[k], labels=test_results[k], ident=k)
            )
            for k in test_smiles
        ]
        torch.save(test_data, os.path.join(self.processed_dir, f"test.pt"))

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if self._setup_data_flag != 1:
            return

        self._setup_data_flag += 1
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._set_processed_data_props()

    def _load_dict(self, input_file_path: str) -> Generator[Dict, None, None]:
        """Loads data from a CSV file as a generator.

        Args:
            input_file_path (str): Path to the CSV file.

        Yields:
            Generator[Dict, None, None]: Generator of data dictionaries.
        """
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                smiles = row["smiles"]
                labels = [
                    bool(int(l)) if l else None for l in (row[k] for k in self.HEADERS)
                ]
                yield dict(features=smiles, labels=labels, ident=row["mol_id"])


class Tox21ChallengeChem(Tox21Challenge):
    """Chemical data reader for Tox21Challenge dataset."""

    READER = dr.ChemDataReader


class Tox21MolNetChem(Tox21MolNet):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader



__all__ = [
    "PubchemBPE",
    "PubChemTokens",
    "SWJSelfies",
    "SWJPreChem",
    "SWJBPE",
    "SWJChem",
]

import gzip
import os
import random
import shutil
import tempfile
import time
from datetime import datetime
from typing import Generator, List, Optional, Tuple, Type

import numpy as np
import pandas as pd
import requests
import torch
import tqdm
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem
from scipy import spatial
from sklearn.cluster import KMeans
from sklearn.model_selection import train_test_split

from chebai.preprocessing import reader as dr
from chebai.preprocessing.datasets.base import DataLoader, XYBaseDataModule
from chebai.preprocessing.datasets.chebi import (
    ChEBIOver50,
    ChEBIOver100,
    ChEBIOverX,
    _ChEBIDataExtractor,
)


class PubChem(XYBaseDataModule):
    """
    Dataset module for PubChem compounds.
    """

    SMILES_INDEX = 0
    LABEL_INDEX = 1
    FULL = 0
    UNLABELED = True
    READER = dr.ChemDataReader

    def __init__(self, *args, k: Optional[int] = 100000, **kwargs):
        """
        Args:
            k (Optional[int]): Number of samples to use. Set to `PubChem.FULL` for full dataset.
            *args: Additional arguments for superclass initialization.
            **kwargs: Additional keyword arguments for superclass initialization.
        """
        self._k = k
        current_year = datetime.today().year
        current_month = datetime.today().month
        self.pubchem_url = f"https://ftp.ncbi.nlm.nih.gov/pubchem/Compound/Monthly/{current_year}-{current_month:02d}-01/Extras/CID-SMILES.gz"

        super(PubChem, self).__init__(*args, **kwargs)

    @property
    def _name(self) -> str:
        """
        Returns:
            str: Name of the dataset.
        """
        return f"Pubchem"

    @property
    def identifier(self) -> tuple:
        """
        Returns:
            tuple: Tuple containing reader name and split label.
        """
        return self.reader.name(), self.split_label

    @property
    def split_label(self) -> str:
        """
        Returns:
            str: Label indicating the split of the dataset ('full' or a specific number).
        """
        if self._k and self._k != self.FULL:
            return str(self._k)
        else:
            return "full"

    @property
    def raw_dir(self) -> str:
        """
        Returns:
            str: Directory path where raw data is stored.
        """
        return os.path.join(self.base_dir, "raw", self.split_label)

    @staticmethod
    def _load_dict(input_file_path: str) -> Generator[dict, None, None]:
        """
        Args:
            input_file_path (str): Path to the input file.

        Yields:
            dict: Dictionary containing 'features', 'labels' (None), and 'ident' fields.
        """
        with open(input_file_path, "r") as input_file:
            for row in input_file:
                ident, smiles = row.split("\t")
                yield dict(features=smiles, labels=None, ident=ident)

    def download(self):
        """
        Downloads PubChem data based on `_k` parameter.
        """
        if not os.path.isfile(os.path.join(self.raw_dir, "smiles.txt")):
            if self._k == PubChem.FULL:
                print("Download from", self.pubchem_url)
                r = requests.get(self.pubchem_url, allow_redirects=True)
                with tempfile.NamedTemporaryFile() as tf:
                    tf.write(r.content)
                    print("Unpacking...")
                    tf.seek(0)
                    with gzip.open(tf, "rb") as f_in:
                        with open(
                            os.path.join(self.raw_dir, "smiles.txt"), "wb"
                        ) as f_out:
                            shutil.copyfileobj(f_in, f_out)
            else:
                full_dataset = self.__class__(k=PubChem.FULL)
                full_dataset.download()
                with open(
                    os.path.join(full_dataset.raw_dir, "smiles.txt"), "r"
                ) as f_in:
                    lines = sum(1 for _ in f_in)
                    selected = frozenset(random.sample(list(range(lines)), k=self._k))
                    f_in.seek(0)
                    selected_lines = list(
                        filter(
                            lambda x: x[0] in selected,
                            enumerate(tqdm.tqdm(f_in, total=lines)),
                        )
                    )
                with open(os.path.join(self.raw_dir, "smiles.txt"), "w") as f_out:
                    f_out.writelines([l for i, l in selected_lines])

    def setup_processed(self):
        """
        Prepares processed data and saves them as Torch tensors.
        """
        filename = os.path.join(self.raw_dir, self.raw_file_names[0])
        print("Load data from file", filename)
        data = self._load_data_from_file(filename)
        print("Create splits")
        train, test = train_test_split(data, train_size=self.train_split)
        del data
        test, val = train_test_split(test, train_size=self.train_split)
        torch.save(train, os.path.join(self.processed_dir, f"train.pt"))
        torch.save(test, os.path.join(self.processed_dir, f"test.pt"))
        torch.save(val, os.path.join(self.processed_dir, f"validation.pt"))

        self.reader.on_finish()

    @property
    def raw_file_names(self) -> List[str]:
        """
        Returns:
            List[str]: List of raw data file names.
        """
        return ["smiles.txt"]

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns:
            List[str]: List of processed data file names.
        """
        return ["test.pt", "train.pt", "validation.pt"]

    def _perform_data_preparation(self, *args, **kwargs):
        """
        Checks for raw data and downloads if necessary.
        """
        print("Check for raw data in", self.raw_dir)
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            print("Downloading data. This may take some time...")
            self.download()
            print("Done")


class PubChemDissimilar(PubChem):
    """
    Subset of PubChem, but choosing the most dissimilar molecules (according to fingerprint)
    """

    def __init__(
        self,
        *args,
        k: Optional[int] = 100000,
        n_random_subsets: Optional[int] = 100,
        random_size_factor: Optional[int] = 5,
        **kwargs,
    ):
        """
        Args:
            k (Optional[int]): Number of entries in this dataset.
            n_random_subsets (Optional[int]): Number of subsets of random data to draw most dissimilar molecules from.
            random_size_factor (Optional[int]): Size of random subsets in relation to k.
            *args: Additional arguments for superclass initialization.
            **kwargs: Additional keyword arguments for superclass initialization.
        """
        self.n_random_subsets = n_random_subsets
        self.random_size_factor = random_size_factor
        super(PubChemDissimilar, self).__init__(*args, k=k, **kwargs)

    @property
    def _name(self) -> str:
        """
        Returns:
            str: Name of the dataset.
        """
        return f"PubchemDissimilar"

    def download(self):
        """
        Downloads the PubChemDissimilar dataset.

        If `k` is set to `PubChem.FULL`, downloads the full dataset.
        Otherwise, generates random subsets to select the most dissimilar molecules based on fingerprints.
        """
        if self._k == PubChem.FULL:
            super().download()
        else:
            # split random subset into n parts, from each part, select the most dissimilar entities
            random_dataset = PubChem(k=self._k * self.random_size_factor)
            random_dataset.download()

            with open(os.path.join(random_dataset.raw_dir, "smiles.txt"), "r") as f_in:
                random_smiles = [
                    [x.strip() for x in s.split("\t")] for s in f_in.readlines()
                ]
                fpgen = AllChem.GetRDKitFPGenerator()
                selected_smiles = []
                print(f"Selecting most dissimilar values from random subsets...")
                for i in tqdm.tqdm(range(self.n_random_subsets)):
                    smiles_i = random_smiles[
                        i
                        * len(random_smiles)
                        // self.n_random_subsets : (i + 1)
                        * len(random_smiles)
                        // self.n_random_subsets
                    ]
                    mols_i = [Chem.MolFromSmiles(smiles) for _, smiles in smiles_i]
                    fps = [
                        fpgen.GetFingerprint(m) if m is not None else m for m in mols_i
                    ]
                    nonnull_fps = [fp for fp in fps if fp is not None]
                    similarity = []
                    for i, fp in enumerate(fps):
                        try:
                            if fp is not None:
                                bulk = DataStructs.BulkTanimotoSimilarity(
                                    fp, nonnull_fps
                                )
                                similarity.append(sum(bulk))
                            else:
                                similarity.append(len(smiles_i))
                        except Exception as e:
                            print(i, smiles_i[i])
                            print(e.with_traceback(None))
                            similarity.append(len(smiles_i))

                    similarity = sorted(zip(smiles_i, similarity), key=lambda x: x[1])
                    selected_smiles += list(
                        list(
                            zip(*similarity[: len(smiles_i) // self.random_size_factor])
                        )[0]
                    )
            with open(os.path.join(self.raw_dir, "smiles.txt"), "w") as f_out:
                f_out.writelines(
                    "\n".join(["\t".join(smiles) for smiles in selected_smiles])
                )


class PubChemKMeans(PubChem):
    """
    Dataset class representing a subset of PubChem dataset clustered using K-Means algorithm.
    The idea is to create distinct distributions where pretraining and test sets are formed from dissimilar data.
    """

    def __init__(
        self,
        *args,
        n_clusters: int = 10000,
        random_size: int = 1000000,
        exclude_data_from: _ChEBIDataExtractor = None,
        validation_size_limit: int = 4000,
        include_min_n_clusters: int = 100,
        **kwargs,
    ):
        """
        Args:
            n_clusters (int): Number of clusters to create using K-Means.
            random_size (int): Size of random dataset to download.
            exclude_data_from (_ChEBIDataExtractor): Dataset which should not overlap with selected clusters
            (remove all clusters that contain data from this dataset).
            validation_size_limit (int): Validation set will contain at most this number of instances.
            include_min_n_clusters (int): Minimum number of clusters to keep if there are not enough clusters that don't
            overlap with the `exclude_data_from` dataset.

            *args: Additional arguments for superclass initialization.
            **kwargs: Additional keyword arguments for superclass initialization.
        """
        self.n_clusters = int(n_clusters)
        self.exclude_data_from = exclude_data_from
        self.validation_size_limit = validation_size_limit
        self.include_min_n_clusters = include_min_n_clusters
        super(PubChemKMeans, self).__init__(*args, k=int(random_size), **kwargs)
        self._fingerprints = None
        self._cluster_centers = None
        self._fingerprints_clustered = None
        self._exclusion_data_clustered = None
        self._cluster_centers_superclustered = None

    @property
    def _name(self) -> str:
        """
        Returns:
            str: Name of the dataset.
        """
        return f"PubchemKMeans"

    @property
    def split_label(self) -> str:
        """
        Returns:
            str: Label describing the split based on number of clusters.
        """
        if self._k and self._k != self.FULL:
            return f"{self.n_clusters}_centers_out_of_{self._k}"
        else:
            return f"{self.n_clusters}_centers_out_of_full"

    @property
    def raw_file_names(self) -> List[str]:
        """
        Clusters generated by K-Means, sorted by size (cluster0 is the largest).
        cluster0 is the training cluster (will be split into train/val/test in processed, used for pretraining)
        Returns:
            List[str]: List of raw file names expected in the raw directory.
        """
        return ["cluster0.txt", "cluster1.txt", "cluster2.txt"]

    @property
    def fingerprints(self) -> pd.DataFrame:
        """
        Creates random dataset, sanitises, creates Mol objects, generates fingerprints (RDKit)
        Saves `fingerprints_df` to `fingerprints.pkl`

        Returns:
            pd.DataFrame: DataFrame containing SMILES and corresponding fingerprints.
        """
        if self._fingerprints is None:
            fingerprints_path = os.path.join(self.raw_dir, "fingerprints.pkl")
            if not os.path.exists(fingerprints_path):
                print(f"No fingerprints found...")
                print(f"Loading random dataset (size: {self._k})...")
                random_dataset = PubChem(k=self._k)
                random_dataset.download()
                with open(
                    os.path.join(random_dataset.raw_dir, "smiles.txt"), "r"
                ) as f_in:
                    random_smiles = [s.split("\t")[1].strip() for s in f_in.readlines()]
                    fpgen = AllChem.GetRDKitFPGenerator()
                    print(f"Converting SMILES to molecules...")
                    mols = [Chem.MolFromSmiles(s) for s in tqdm.tqdm(random_smiles)]
                    print(f"Generating Fingerprints...")
                    fps = [
                        fpgen.GetFingerprint(m) if m is not None else m
                        for m in tqdm.tqdm(mols)
                    ]
                    d = {"smiles": random_smiles, "fps": fps}
                    fingerprints_df = pd.DataFrame(d, columns=["smiles", "fps"])
                    fingerprints_df = fingerprints_df.dropna()
                    fingerprints_df.to_pickle(open(fingerprints_path, "wb"))
                    self._fingerprints = fingerprints_df
            else:
                self._fingerprints = pd.read_pickle(open(fingerprints_path, "rb"))
        return self._fingerprints

    def _build_clusters(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Performs K-Means clustering on fingerprints and saves cluster information.

        Returns:
            tuple: Tuple containing cluster centers DataFrame and clustered fingerprints DataFrame.
        """
        fingerprints_clustered_path = os.path.join(
            self.raw_dir, "fingerprints_clustered.pkl"
        )
        cluster_centers_path = os.path.join(self.raw_dir, f"cluster_centers.pkl")
        print(f"Starting k-means clustering...")
        start_time = time.perf_counter()
        kmeans = KMeans(n_clusters=self.n_clusters, random_state=0, n_init="auto")
        fps = np.array([list(vec) for vec in self.fingerprints["fps"].tolist()])
        kmeans.fit(fps)
        print(f"Finished k-means in {time.perf_counter() - start_time:.2f} seconds")
        fingerprints_df = self.fingerprints
        fingerprints_df["label"] = kmeans.labels_
        fingerprints_df.to_pickle(
            open(
                fingerprints_clustered_path,
                "wb",
            )
        )
        cluster_df = pd.DataFrame(
            data={"centers": [center for center in kmeans.cluster_centers_]}
        )
        cluster_df.to_pickle(
            open(
                cluster_centers_path,
                "wb",
            )
        )

        return cluster_df, fingerprints_df

    def _exclude_clusters(self, cluster_centers: pd.DataFrame) -> pd.DataFrame:
        """
        Excludes clusters based on data from an exclusion dataset (in a training setup, this is the labeled dataset,
        usually ChEBI). The goal is to avoid having similar data in the labeled training and the PubChem evaluation.

         Loads data from `exclude_data_from` dataset, generates mols, fingerprints, finds closest cluster centre for
         each fingerprint, saves data to `exclusion_data_clustered.pkl`, returns all clusters with no instances from the
         exclusion data (or the n clusters with the lowest number of instances if there are less than n clusters with no
         instances, n being the minimum number of clusters to include)

        Args:
            cluster_centers (pd.DataFrame): DataFrame of cluster centers.

        Returns:
            pd.DataFrame: DataFrame of filtered cluster centers.
        """
        exclusion_data_path = os.path.join(self.raw_dir, "exclusion_data_clustered.pkl")
        cluster_centers_np = np.array(
            [
                [cci for cci in cluster_center]
                for cluster_center in cluster_centers["centers"]
            ]
        )
        if self.exclude_data_from is not None:
            if not os.path.exists(exclusion_data_path):
                print(f"Loading data for exclusion of clusters...")
                raw_chebi = []
                for filename in self.exclude_data_from.raw_file_names:
                    raw_chebi.append(
                        pd.read_pickle(
                            open(
                                os.path.join(self.exclude_data_from.raw_dir, filename),
                                "rb",
                            )
                        )
                    )
                raw_chebi = pd.concat(raw_chebi)
                raw_chebi_smiles = np.array(raw_chebi["SMILES"])
                fpgen = AllChem.GetRDKitFPGenerator()
                print(f"Converting SMILES to molecules...")
                mols = [Chem.MolFromSmiles(s) for s in tqdm.tqdm(raw_chebi_smiles)]
                print(f"Generating Fingerprints...")
                chebi_fps = [
                    fpgen.GetFingerprint(m) if m is not None else m
                    for m in tqdm.tqdm(mols)
                ]
                print(f"Finding cluster for each instance from exclusion-data")
                chebi_fps = np.array([list(fp) for fp in chebi_fps if fp is not None])
                tree = spatial.KDTree(cluster_centers_np)
                chebi_clusters = [tree.query(fp)[1] for fp in chebi_fps]
                chebi_clusters_df = pd.DataFrame(
                    {"fp": [fp for fp in chebi_fps], "center_id": chebi_clusters},
                    columns=["fp", "center_id"],
                )
                chebi_clusters_df.to_pickle(open(exclusion_data_path, "wb"))
            else:
                chebi_clusters_df = pd.read_pickle(open(exclusion_data_path, "rb"))
            # filter pubchem clusters and remove all that contain data from the exclusion set
            print(f"Removing clusters with data from exclusion-set")
            counts = chebi_clusters_df["center_id"].value_counts()
            cluster_centers["n_chebi_instances"] = counts
            cluster_centers["n_chebi_instances"].fillna(0, inplace=True)
            cluster_centers.sort_values(
                by="n_chebi_instances", ascending=False, inplace=True
            )
            zero_centers = cluster_centers[cluster_centers["n_chebi_instances"] == 0]
            if len(zero_centers) > self.include_min_n_clusters:
                cluster_centers = zero_centers
            else:
                cluster_centers = cluster_centers[-self.include_min_n_clusters :]
        return cluster_centers

    @property
    def cluster_centers(self) -> pd.DataFrame:
        """
        Loads cluster centers from file if possible, otherwise calls `self._build_clusters()`.
        Returns:
            pd.DataFrame: DataFrame of cluster centers.
        """
        cluster_centers_path = os.path.join(self.raw_dir, f"cluster_centers.pkl")
        if self._cluster_centers is None:
            if os.path.exists(cluster_centers_path):
                self._cluster_centers = pd.read_pickle(open(cluster_centers_path, "rb"))
            else:
                self._cluster_centers = self._build_clusters()[0]
        return self._cluster_centers

    @property
    def fingerprints_clustered(self) -> pd.DataFrame:
        """
        Loads fingerprints with assigned clusters from file if possible, otherwise calls `self._build_clusters()`.
        Returns:
            pd.DataFrame: DataFrame of clustered fingerprints.
        """
        fingerprints_path = os.path.join(self.raw_dir, f"fingerprints_clustered.pkl")
        if self._fingerprints_clustered is None:
            if os.path.exists(fingerprints_path):
                self._fingerprints_clustered = pd.read_pickle(
                    open(fingerprints_path, "rb")
                )
            else:
                self._fingerprints_clustered = self._build_clusters()[1]
        return self._fingerprints_clustered

    @property
    def cluster_centers_superclustered(self) -> pd.DataFrame:
        """
        Calls `_exclude_clusters()` which removes all clusters that contain data from the exclusion set (usually the
        ChEBI, i.e., the labeled dataset).
        Runs KMeans with 3 clusters on remaining data, saves cluster centres with assigned supercluster-labels to
        `cluster_centers_superclustered.pkl`
        Returns:
            pd.DataFrame: DataFrame of superclustered cluster centers.
        """
        cluster_centers_path = os.path.join(
            self.raw_dir, f"cluster_centers_superclustered.pkl"
        )
        if self._cluster_centers_superclustered is None:
            if not os.path.exists(cluster_centers_path):
                clusters_filtered = self._exclude_clusters(self.cluster_centers)
                print(f"Superclustering PubChem clusters")
                kmeans = KMeans(n_clusters=3, random_state=0, n_init="auto")
                clusters_np = np.array(
                    [[cci for cci in center] for center in clusters_filtered["centers"]]
                )
                kmeans.fit(clusters_np)
                clusters_filtered["label"] = kmeans.labels_
                clusters_filtered.to_pickle(
                    open(
                        os.path.join(
                            self.raw_dir, "cluster_centers_superclustered.pkl"
                        ),
                        "wb",
                    )
                )
                self._cluster_centers_superclustered = clusters_filtered
            else:
                self._cluster_centers_superclustered = pd.read_pickle(
                    open(
                        os.path.join(
                            self.raw_dir, f"cluster_centers_superclustered.pkl"
                        ),
                        "rb",
                    )
                )
        return self._cluster_centers_superclustered

    def download(self):
        """
        Downloads the PubChemKMeans dataset. This function creates the complete dataset (including train, test, and
        validation splits). Most of the steps are hidden in properties (e.g., `self.fingerprints_clustered` triggers
        the download of a random dataset, the calculation of fingerprints for it and the KMeans clustering)
        The final splits are created by assigning all fingerprints that belong to a cluster of a certain supercluster
        to a dataset. This creates 3 datasets (for each of the 3 superclusters), the datasets are saved as validation,
        test and train based on their size. The validation set is limited to `self.validation_size_limit` entries.
        """
        if self._k == PubChem.FULL:
            super().download()
        else:
            if not all(
                os.path.exists(os.path.join(self.raw_dir, file))
                for file in self.raw_file_names
            ):
                fingerprints = self.fingerprints_clustered
                fingerprints["big_cluster_assignment"] = fingerprints["label"].apply(
                    lambda l: (
                        -1
                        if l not in self.cluster_centers_superclustered.index
                        else self.cluster_centers_superclustered.loc[int(l), "label"]
                    )
                )
                fp_grouped = fingerprints.groupby("big_cluster_assignment")
                splits = [fp_grouped.get_group(g) for g in fp_grouped.groups if g != -1]
                splits[0] = splits[0][: self.validation_size_limit]
                splits.sort(key=lambda x: len(x))
                for i, name in enumerate(["cluster2", "cluster1", "cluster0"]):
                    if not os.path.exists(os.path.join(self.raw_dir, f"{name}.txt")):
                        open(os.path.join(self.raw_dir, f"{name}.txt"), "x").close()
                    with open(os.path.join(self.raw_dir, f"{name}.txt"), "w") as f:
                        for id, row in splits[i].iterrows():
                            f.writelines(f"{id}\t{row['smiles']}\n")


class PubChemDissimilarSMILES(PubChemDissimilar):
    """
    Subset of PubChem, selecting most dissimilar molecules based on fingerprints.

    Inherits from PubChemDissimilar.

    Attributes:
        READER (type): Data reader type for chemical data.
    """

    READER: Type[dr.ChemDataReader] = dr.ChemDataReader


class SWJPreChem(PubChem):
    """
    Subset of PubChem with unlabeled data, specific to SWJPre.

    Inherits from PubChem.

    Attributes:
        UNLABELED (bool): Indicates if the data is unlabeled.
        _name (str): Name of the dataset.
    """

    UNLABELED: bool = True

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.
        """
        return f"SWJpre"

    def download(self):
        """
        Raises an exception since required raw files are not found.
        """
        raise Exception("Required raw files not found")

    @property
    def identifier(self) -> Tuple[str]:
        """
        Returns the identifier for the dataset.
        """
        return (self.reader.name(),)

    @property
    def raw_dir(self) -> str:
        """
        Returns the directory path for raw data.
        """
        return os.path.join("data", self._name, "raw")


class SWJSelfies(SWJPreChem):
    """
    Subset of SWJPreChem using SelfiesReader for data reading.

    Inherits from SWJPreChem.

    Attributes:
        READER (type): Data reader type for chemical data (SelfiesReader).
    """

    READER: Type[dr.SelfiesReader] = dr.SelfiesReader


class PubchemChem(PubChem):
    """
    Subset of PubChem using ChemDataReader for data reading.

    Inherits from PubChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataReader).
    """

    READER: Type[dr.ChemDataReader] = dr.ChemDataReader


class PubchemBPE(PubChem):
    """
    Subset of PubChem using ChemBPEReader for data reading.

    Inherits from PubChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemBPEReader).
    """

    READER: Type[dr.ChemBPEReader] = dr.ChemBPEReader


class SWJChem(SWJPreChem):
    """
    Subset of SWJPreChem using ChemDataUnlabeledReader for data reading.

    Inherits from SWJPreChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataUnlabeledReader).
    """

    READER: Type[dr.ChemDataUnlabeledReader] = dr.ChemDataUnlabeledReader


class SWJBPE(SWJPreChem):
    """
    Subset of SWJPreChem using ChemBPEReader for data reading.

    Inherits from SWJPreChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemBPEReader).
    """

    READER: Type[dr.ChemBPEReader] = dr.ChemBPEReader


class PubChemTokens(PubChem):
    """
    Subset of PubChem using ChemDataReader for data reading.

    Inherits from PubChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataReader).
    """

    READER: Type[dr.ChemDataReader] = dr.ChemDataReader


class Hazardous(SWJChem):
    """
    Subset of SWJChem for hazardous compounds.

    Inherits from SWJChem.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataUnlabeledReader).
    """

    READER: Type[dr.ChemDataUnlabeledReader] = dr.ChemDataUnlabeledReader

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.
        """
        return f"PubChemHazardous"

    def setup_processed(self):
        """
        Sets up the processed data.
        """
        filename = os.path.join(self.raw_dir, self.raw_file_names[0])
        print("Load data from file", filename)
        data = self._load_data_from_file(filename)
        torch.save(data, os.path.join(self.processed_dir, f"all.pt"))

        self.reader.on_finish()

    def processed_file_names(self) -> List[str]:
        """
        Returns the list of processed file names.
        """
        return ["all.pt"]

    def download(self):
        """
        Downloads hazardous compound data from PubChem.
        """
        # requires the / a hazardous subset from pubchem, e.g. obtained by entering
        # "PubChem: PubChem Compound TOC: GHS Classification" in the pubchem search -> download -> csv
        csv_path = os.path.join(self.raw_dir, "pubchem_hazardous_compound_list.csv")
        compounds = pd.read_csv(csv_path)
        smiles_list = []
        for id, compound in compounds.iterrows():
            if (
                not isinstance(compound["cmpdsynonym"], str)
                or "CHEBI" not in compound["cmpdsynonym"]
            ):
                smiles_list.append(f"{compound['cid']}\t{compound['isosmiles']}")
        with open(os.path.join(self.raw_dir, "smiles.txt"), "w") as f:
            f.write("\n".join(smiles_list))


class SWJPreChem(PubChem):
    """
    Subset of PubChem specific to SWJpre with unlabeled data.

    Inherits from PubChem.

    Attributes:
        UNLABELED (bool): Indicates if the data is unlabeled.
        _name (str): Name of the dataset.
    """

    UNLABELED: bool = True

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.

        Returns:
            str: Name of the dataset.
        """
        return f"SWJpre"

    def download(self) -> None:
        """
        Raises an exception since required raw files are not found.

        Raises:
            Exception: If required raw files are not found.
        """
        raise Exception("Required raw files not found")

    @property
    def identifier(self) -> Tuple[str]:
        """
        Returns the identifier for the dataset.

        Returns:
            tuple: A tuple containing the name of the reader.
        """
        return (self.reader.name(),)


class LabeledUnlabeledMixed(XYBaseDataModule):
    """
    Mixed dataset combining labeled and unlabeled data.

    Inherits from XYBaseDataModule.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataReader).
    """

    READER: Type[dr.ChemDataReader] = dr.ChemDataReader

    def __init__(
        self, labeled: XYBaseDataModule, unlabeled: XYBaseDataModule, *args, **kwargs
    ):
        """
        Args:
            labeled (XYBaseDataModule): Labeled dataset module.
            unlabeled (XYBaseDataModule): Unlabeled dataset module.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        self.labeled = labeled
        self.unlabeled = unlabeled
        super().__init__(*args, **kwargs)

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.
        """
        return f"Mixed_{self.labeled._name}_{self.unlabeled._name}"

    def dataloader(self, kind: str, **kwargs) -> DataLoader:
        """
        Returns a dataloader for the specified kind.

        Args:
            kind (str): Type of data (e.g., 'train', 'validation', 'test').
            **kwargs: Additional keyword arguments for DataLoader.

        Returns:
            DataLoader: DataLoader instance.
        """
        labeled_data = torch.load(
            os.path.join(self.labeled.processed_dir, f"{kind}.pt"), weights_only=False
        )
        unlabeled_data = torch.load(
            os.path.join(self.unlabeled.processed_dir, f"{kind}.pt"), weights_only=False
        )
        if self.data_limit is not None:
            labeled_data = labeled_data[: self.data_limit]
            unlabeled_data = unlabeled_data[: self.data_limit]
        return DataLoader(
            labeled_data + unlabeled_data,
            collate_fn=self.reader.collator,
            batch_size=self.batch_size,
            **kwargs,
        )

    @property
    def raw_file_names(self) -> List[str]:
        """
        Returns the list of raw file names (empty for mixed dataset).
        """
        return []

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns the list of processed file names.
        """
        return ["test.pt", "train.pt", "validation.pt"]

    def setup_processed(self):
        """
        Sets up the processed data by setting up labeled and unlabeled datasets.
        """
        self.labeled.setup()
        self.unlabeled.setup()


class PubToxAndChebiX(LabeledUnlabeledMixed):
    """
    Mixed dataset combining PubChem and ChEBI datasets.

    Inherits from LabeledUnlabeledMixed.

    Attributes:
        READER (type): Data reader type for chemical data (ChemDataReader).
        CHEBI_X (type): Specific ChEBI dataset type.
    """

    READER: Type[dr.ChemDataReader] = dr.ChemDataReader
    CHEBI_X: Type[ChEBIOverX] = ChEBIOverX

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(
            self.CHEBI_X(*args, **kwargs), PubchemChem(*args, **kwargs), *args, **kwargs
        )

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.
        """
        return "PubToxU" + self.labeled._name


class PubToxAndChebi100(PubToxAndChebiX):
    """
    Mixed dataset combining PubChem and ChEBI datasets with over 100 entries.

    Inherits from PubToxAndChebiX.

    Attributes:
        CHEBI_X (type): ChEBI dataset with over 100 entries.
    """

    CHEBI_X: Type[ChEBIOver100] = ChEBIOver100


class PubToxAndChebi50(PubToxAndChebiX):
    """
    Mixed dataset combining PubChem and ChEBI datasets with over 50 entries.

    Inherits from PubToxAndChebiX.

    Attributes:
        CHEBI_X (type): ChEBI dataset with over 50 entries.
    """

    CHEBI_X: Type[ChEBIOver50] = ChEBIOver50


class PubChemDeepSMILES(PubChem):
    """
    Subset of PubChem using DeepChemDataReader for data reading.

    Inherits from PubChem.

    Attributes:
        READER (type): Data reader type for chemical data (DeepChemDataReader).
    """

    READER: Type[dr.DeepChemDataReader] = dr.DeepChemDataReader


class PubChemSELFIES(PubChem):
    """
    Subset of PubChem using SelfiesReader for data reading.

    Inherits from PubChem.

    Attributes:
        READER (type): Data reader type for chemical data (SelfiesReader).
    """

    READER: Type[dr.SelfiesReader] = dr.SelfiesReader



import unittest
from unittest.mock import mock_open, patch

import networkx as nx

from chebai.preprocessing.datasets.chebi import ChEBIOverXPartial
from tests.unit.mock_data.ontology_mock_data import ChebiMockOntology


class TestChEBIOverX(unittest.TestCase):

    @classmethod
    @patch.multiple(ChEBIOverXPartial, __abstractmethods__=frozenset())
    @patch("os.makedirs", return_value=None)
    def setUpClass(cls, mock_makedirs) -> None:
        """
        Set up the ChEBIOverXPartial instance with a mock processed directory path and a test graph.
        """
        cls.chebi_extractor = ChEBIOverXPartial(top_class_id=11111, chebi_version=231)
        cls.test_graph = ChebiMockOntology.get_transitively_closed_graph()

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=ChebiMockOntology.get_raw_data(),
    )
    def test_extract_class_hierarchy(self, mock_open: mock_open) -> None:
        """
        Test the extraction of class hierarchy and validate the structure of the resulting graph.
        """
        # Mock the output of fastobo.loads
        self.chebi_extractor.top_class_id = 11111
        graph: nx.DiGraph = self.chebi_extractor._extract_class_hierarchy("fake_path")

        # Validate the graph structure
        self.assertIsInstance(
            graph, nx.DiGraph, "The result should be a directed graph."
        )

        # Check nodes
        expected_nodes = {11111, 54321, 12345, 99999}
        expected_edges = {
            (54321, 12345),
            (54321, 99999),
            (11111, 54321),
            (11111, 12345),
            (11111, 99999),
            (12345, 99999),
        }
        self.assertEqual(
            set(graph.nodes),
            expected_nodes,
            f"The graph nodes do not match the expected nodes for top class {self.chebi_extractor.top_class_id} hierarchy.",
        )

        # Check edges
        self.assertEqual(
            expected_edges,
            set(graph.edges),
            "The graph edges do not match the expected edges.",
        )

        # Check number of nodes and edges
        self.assertEqual(
            len(graph.nodes),
            len(expected_nodes),
            "The number of nodes should match the actual number of nodes in the graph.",
        )

        self.assertEqual(
            len(expected_edges),
            len(graph.edges),
            "The number of transitive edges should match the actual number of transitive edges in the graph.",
        )

        self.chebi_extractor.top_class_id = 22222
        graph = self.chebi_extractor._extract_class_hierarchy("fake_path")

        # Check nodes with top class as 22222
        self.assertEqual(
            set(graph.nodes),
            {67890, 88888, 12345, 99999, 22222},
            f"The graph nodes do not match the expected nodes for top class {self.chebi_extractor.top_class_id} hierarchy.",
        )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=ChebiMockOntology.get_raw_data(),
    )
    def test_extract_class_hierarchy_with_bottom_cls(
        self, mock_open: mock_open
    ) -> None:
        """
        Test the extraction of class hierarchy and validate the structure of the resulting graph.
        """
        self.chebi_extractor.top_class_id = 88888
        graph: nx.DiGraph = self.chebi_extractor._extract_class_hierarchy("fake_path")

        # Check nodes with top class as 88888
        self.assertEqual(
            set(graph.nodes),
            {self.chebi_extractor.top_class_id},
            f"The graph nodes do not match the expected nodes for top class {self.chebi_extractor.top_class_id} hierarchy.",
        )

    @patch("pandas.DataFrame.to_csv")
    @patch("pandas.read_pickle")
    @patch.object(ChEBIOverXPartial, "_get_data_size", return_value=4.0)
    @patch("torch.load")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=ChebiMockOntology.get_raw_data(),
    )
    def test_single_label_data_split(
        self, mock_open, mock_load, mock_get_data_size, mock_read_pickle, mock_to_csv
    ) -> None:
        """
        Test the single-label data splitting functionality of the ChebiExtractor class.

        This test mocks several key methods (file operations, torch loading, and pandas functions)
        to ensure that the class hierarchy is properly extracted, data is processed into a raw dataset,
        and the data splitting logic works as intended without actual file I/O.

        It also verifies that there is no overlap between training, validation, and test sets.
        """
        self.chebi_extractor.top_class_id = 11111
        self.chebi_extractor.THRESHOLD = 3
        self.chebi_extractor.chebi_version_train = None

        graph: nx.DiGraph = self.chebi_extractor._extract_class_hierarchy("fake_path")
        data_df = self.chebi_extractor._graph_to_raw_dataset(graph)

        mock_read_pickle.return_value = data_df
        data_pt = self.chebi_extractor._load_data_from_file("fake/path")

        # Verify that the data contains only 1 label
        self.assertEqual(len(data_pt[0]["labels"]), 1)

        mock_load.return_value = data_pt

        # Retrieve the data splits (train, validation, and test)
        train_split = self.chebi_extractor.dynamic_split_dfs["train"]
        validation_split = self.chebi_extractor.dynamic_split_dfs["validation"]
        test_split = self.chebi_extractor.dynamic_split_dfs["test"]

        train_idents = set(train_split["ident"])
        val_idents = set(validation_split["ident"])
        test_idents = set(test_split["ident"])

        # Ensure there is no overlap between train and test sets
        self.assertEqual(
            len(train_idents.intersection(test_idents)),
            0,
            "Train and test sets should not overlap.",
        )

        # Ensure there is no overlap between validation and test sets
        self.assertEqual(
            len(val_idents.intersection(test_idents)),
            0,
            "Validation and test sets should not overlap.",
        )

        # Ensure there is no overlap between train and validation sets
        self.assertEqual(
            len(train_idents.intersection(val_idents)),
            0,
            "Train and validation sets should not overlap.",
        )


if __name__ == "__main__":
    unittest.main()



import torch


class ElectraPreLoss(torch.nn.Module):
    """
    Custom loss module for pre-training ELECTRA-like models.

    This module computes a combined loss from two CrossEntropyLosses:
    one for generator predictions and another for discriminator predictions.

    Attributes:
        ce (torch.nn.CrossEntropyLoss): Cross entropy loss function.

    Methods:
        forward(input, target, **loss_kwargs):
            Computes the combined loss for generator and discriminator predictions.

    """

    def __init__(self):
        """
        Initializes the ElectraPreLoss module.
        """
        super().__init__()
        self.ce = torch.nn.CrossEntropyLoss()

    def forward(self, input, target, **loss_kwargs):
        """
        Forward pass for computing the combined loss.

        Args:
            input (tuple): A tuple containing generator predictions (gen_pred, disc_pred).
            target (tuple): A tuple containing generator targets (gen_tar, disc_tar).
            **loss_kwargs: Additional keyword arguments.

        Returns:
            torch.Tensor: Combined loss of generator and discriminator predictions.
        """
        t, p = input
        gen_pred, disc_pred = t
        gen_tar, disc_tar = p

        # Compute losses for generator and discriminator
        gen_loss = self.ce(target=torch.argmax(gen_tar.int(), dim=-1), input=gen_pred)
        disc_loss = self.ce(
            target=torch.argmax(disc_tar.int(), dim=-1), input=disc_pred
        )
        return gen_loss + disc_loss



import inspect
import os
import sys
from abc import ABC
from itertools import islice
from typing import Any, Dict, List, Optional

import deepsmiles
import selfies as sf
from pysmiles.read_smiles import _tokenize
from transformers import RobertaTokenizerFast

from chebai.preprocessing.collate import DefaultCollator, RaggedCollator

EMBEDDING_OFFSET = 10
PADDING_TOKEN_INDEX = 0
MASK_TOKEN_INDEX = 1
CLS_TOKEN = 2


class DataReader:
    """
    Base class for reading and preprocessing data. Turns the raw input data (e.g., a SMILES string) into the model
    input format (e.g., a list of tokens).

    Args:
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments (not used).
    """

    COLLATOR = DefaultCollator

    def __init__(
        self,
        collator_kwargs: Optional[Dict[str, Any]] = None,
        token_path: Optional[str] = None,
        **kwargs,
    ):
        if collator_kwargs is None:
            collator_kwargs = dict()
        self.collator = self.COLLATOR(**collator_kwargs)
        self.dirname = os.path.dirname(inspect.getfile(self.__class__))
        self._token_path = token_path

    def _get_raw_data(self, row: Dict[str, Any]) -> Any:
        """Get raw data from the row."""
        return row["features"]

    def _get_raw_label(self, row: Dict[str, Any]) -> Any:
        """Get raw label from the row."""
        return row["labels"]

    def _get_raw_id(self, row: Dict[str, Any]) -> Any:
        """Get raw ID from the row."""
        return row.get("ident", row["features"])

    def _get_raw_group(self, row: Dict[str, Any]) -> Any:
        """Get raw group from the row."""
        return row.get("group", None)

    def _get_additional_kwargs(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """Get additional keyword arguments from the row."""
        return row.get("additional_kwargs", dict())

    def name(cls) -> str:
        """Returns the name of the data reader."""
        raise NotImplementedError

    @property
    def token_path(self) -> str:
        """Get token path, create file if it does not exist yet."""
        if self._token_path is not None:
            return self._token_path
        token_path = os.path.join(self.dirname, "bin", self.name(), "tokens.txt")
        os.makedirs(os.path.join(self.dirname, "bin", self.name()), exist_ok=True)
        if not os.path.exists(token_path):
            with open(token_path, "x"):
                pass
        return token_path

    def _read_id(self, raw_data: Any) -> Any:
        """Read and return ID from raw data."""
        return raw_data

    def _read_data(self, raw_data: Any) -> Any:
        """Read and return data from raw data."""
        return raw_data

    def _read_label(self, raw_label: Any) -> Any:
        """Read and return label from raw label."""
        return raw_label

    def _read_group(self, raw: Any) -> Any:
        """Read and return group from raw group data."""
        return raw

    def _read_components(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """Read and return components from the row."""
        return dict(
            features=self._get_raw_data(row),
            labels=self._get_raw_label(row),
            ident=self._get_raw_id(row),
            group=self._get_raw_group(row),
            additional_kwargs=self._get_additional_kwargs(row),
        )

    def to_data(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """Convert raw row data to processed data."""
        d = self._read_components(row)
        return dict(
            features=self._read_data(d["features"]),
            labels=self._read_label(d["labels"]),
            ident=self._read_id(d["ident"]),
            group=self._read_group(d["group"]),
            **d["additional_kwargs"],
        )

    def on_finish(self) -> None:
        """Hook to run at the end of preprocessing."""
        return


class TokenIndexerReader(DataReader, ABC):
    """
    Abstract base class for reading tokenized data and mapping tokens to unique indices.

    This class maintains a cache of token-to-index mappings that can be extended during runtime,
    and saves new tokens to a persistent file at the end of processing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        with open(self.token_path, "r") as pk:
            self.cache: Dict[str, int] = {
                token.strip(): idx for idx, token in enumerate(pk)
            }
        self._loaded_tokens_count = len(self.cache)

    def _get_token_index(self, token: str) -> int:
        """Returns a unique number for each token, automatically adds new tokens."""
        if not str(token) in self.cache:
            self.cache[(str(token))] = len(self.cache)
        return self.cache[str(token)] + EMBEDDING_OFFSET

    def on_finish(self) -> None:
        """
        Saves the current cache of tokens to the token file.This method is called after all data processing is complete.
        """
        print(f"first 10 tokens: {list(islice(self.cache, 10))}")

        total_tokens = len(self.cache)
        if total_tokens > self._loaded_tokens_count:
            print("New tokens added to the cache, Saving them to token file.....")

            assert sys.version_info >= (
                3,
                7,
            ), "This code requires Python 3.7 or higher."
            # For python 3.7+, the standard dict type preserves insertion order, and is iterated over in same order
            # https://docs.python.org/3/whatsnew/3.7.html#summary-release-highlights
            # https://mail.python.org/pipermail/python-dev/2017-December/151283.html
            new_tokens = list(
                islice(self.cache, self._loaded_tokens_count, total_tokens)
            )

            with open(self.token_path, "a") as pk:
                print(f"saving new {len(new_tokens)} tokens to {self.token_path}...")
                pk.writelines([f"{c}\n" for c in new_tokens])


class ChemDataReader(TokenIndexerReader):
    """
    Data reader for chemical data using SMILES tokens.
    """

    COLLATOR = RaggedCollator

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "smiles_token"

    def _read_data(self, raw_data: str) -> List[int]:
        """
        Reads and tokenizes raw SMILES data into a list of token indices.

        Args:
            raw_data (str): The raw SMILES string to be tokenized.

        Returns:
            List[int]: A list of integers representing the indices of the SMILES tokens.
        """
        return [self._get_token_index(v[1]) for v in _tokenize(raw_data)]


class DeepChemDataReader(ChemDataReader):
    """
    Data reader for chemical data using DeepSMILES tokens.

    Args:
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.converter = deepsmiles.Converter(rings=True, branches=True)
        self.error_count = 0

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "deepsmiles_token"

    def _read_data(self, raw_data: str) -> Optional[List[int]]:
        """Read and tokenize raw data using DeepSMILES."""
        try:
            tokenized = _tokenize(self.converter.encode(raw_data))
            tokenized = [self._get_token_index(v[1]) for v in tokenized]
        except ValueError as e:
            print(f"could not process {raw_data}")
            print(f"Corresponding deepSMILES: {self.converter.encode(raw_data)}")
            print(f"\t{e}")
            self.error_count += 1
            print(f"\terror count: {self.error_count}")
            tokenized = None
        return tokenized


class ChemDataUnlabeledReader(ChemDataReader):
    """
    Data reader for unlabeled chemical data using SMILES tokens.

    Args:
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments.
    """

    COLLATOR = RaggedCollator

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "smiles_token_unlabeled"

    def _get_raw_label(self, row: Dict[str, Any]) -> None:
        """Returns None as there are no labels."""
        return None


class ChemBPEReader(DataReader):
    """
    Data reader for chemical data using BPE tokenization.

    Args:
        data_path: Path for the pretrained BPE tokenizer.
        max_len: Maximum length of the tokenized sequence.
        vsize: Vocabulary size for the tokenizer (not used).
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments.
    """

    COLLATOR = RaggedCollator

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "smiles_bpe"

    def __init__(
        self,
        *args,
        data_path: Optional[str] = None,
        max_len: int = 1800,
        vsize: int = 4000,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.tokenizer = RobertaTokenizerFast.from_pretrained(
            data_path, max_len=max_len
        )

    def _get_raw_data(self, row: Dict[str, Any]) -> List[int]:
        """Tokenize raw data using BPE tokenizer."""
        return self.tokenizer(row["features"])["input_ids"]


class SelfiesReader(ChemDataReader):
    """
    Data reader for chemical data using SELFIES tokens.

    Args:
        data_path: Path for the pretrained BPE tokenizer.
        max_len: Maximum length of the tokenized sequence.
        vsize: Vocabulary size for the tokenizer.
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments.
    """

    COLLATOR = RaggedCollator

    def __init__(
        self,
        *args,
        data_path: Optional[str] = None,
        max_len: int = 1800,
        vsize: int = 4000,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.error_count = 0
        sf.set_semantic_constraints("hypervalent")

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "selfies"

    def _read_data(self, raw_data: str) -> Optional[List[int]]:
        """Read and tokenize raw data using SELFIES."""
        try:
            tokenized = sf.split_selfies(sf.encoder(raw_data.strip(), strict=True))
            tokenized = [self._get_token_index(v) for v in tokenized]
        except Exception as e:
            print(f"could not process {raw_data}")
            # print(f'\t{e}')
            self.error_count += 1
            print(f"\terror count: {self.error_count}")
            tokenized = None
            # if self.error_count > 20:
            #    raise Exception('Too many errors')
        return tokenized


class OrdReader(DataReader):
    """
    Data reader that converts characters to their ordinal values.

    Args:
        collator_kwargs: Optional dictionary of keyword arguments for the collator.
        token_path: Optional path for the token file.
        kwargs: Additional keyword arguments.
    """

    COLLATOR = RaggedCollator

    @classmethod
    def name(cls) -> str:
        """Returns the name of the data reader."""
        return "ord"

    def _read_data(self, raw_data: str) -> List[int]:
        """Convert characters in raw data to their ordinal values."""
        return [ord(s) for s in raw_data]
